"""
Cython declarations for the NES cartridges
"""
from .system cimport InterruptListener   # used by MMC3 to generate IRQs


cdef enum:
    BYTES_PER_KB = 1024

cdef enum:
    # cart memory map - note that some memory is connected to CPU bus, others to PPU bus
    RAM_START = 0x6000      # address in the CPU memory space
    PRG_ROM_START = 0x8000  # address in the CPU memory space
    CHR_ROM_START = 0x0000  # address in the PPU memory space


cdef class NESCart:
    """
    NES Cartridge interface
    """
    cdef int nametable_mirror_pattern[4]

    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)
    cdef unsigned char read_ppu(self, int address)
    cdef void write_ppu(self, int address, unsigned char value)

    cdef void irq_tick(self)


### Mapper 0 (aka NROM) ################################################################################################

# would be nicer to have these in a named enum, but that seems to generate Cython compile error "Array dimension not
# integer", so it's not clear what that is ending up doing.  For now, just name the constants with the mapper number.
cdef enum:
    # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
    M0_MAX_CART_RAM_SIZE = 8 * BYTES_PER_KB   # up to 8kB of RAM
    M0_MAX_PRG_ROM_SIZE = 32 * BYTES_PER_KB   # up to 32kB of program ROM
    M0_CHR_MEM_SIZE = 8 * BYTES_PER_KB        # always (?) 8kB of chr memory (can be RAM or always ROM?)


cdef class NESCart0(NESCart):
    """
    Basic NES Cartridge (Type 0 / MMC0)
    Consists of up to 8kB RAM, 32kB PRG ROM, 8kB CHR ROM
    """
    cdef unsigned char ram[M0_MAX_CART_RAM_SIZE]
    cdef unsigned char prg_rom[M0_MAX_PRG_ROM_SIZE]
    cdef unsigned char chr_mem[M0_CHR_MEM_SIZE]
    cdef int prg_start_addr, prg_rom_size, ram_size, chr_mem_writeable, prg_rom_writeable     # cart metadata

    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)
    cdef unsigned char read_ppu(self, int address)
    cdef void write_ppu(self, int address, unsigned char value)


### Mapper 1 (aka MMC1, SxROM) #########################################################################################
# ref: https://wiki.nesdev.com/w/index.php/MMC1

cdef enum:
    # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
    M1_PRG_RAM_BANK_SIZE = 8 * BYTES_PER_KB    # 8kB RAM banks (can be zero of these)
    M1_MAX_PRG_RAM_BANKS = 4                   # can have up to 4 x 8kb RAM banks, but only on SXROM
    M1_PRG_ROM_BANK_SIZE = 16 * BYTES_PER_KB   # Each bank is 16kB of program ROM
    M1_MAX_PRG_BANKS = 16                      # up to 256kB of program ROM (16x16kb) - 512kb MMC1 ROMs not supported
    M1_CHR_ROM_BANK_SIZE = 4 * BYTES_PER_KB    # chr rom in 4kB banks
    M1_MAX_CHR_BANKS = 32                      # up to 128kb of CHR rom in (up to 32 x 4kb)

    # main memory map (read)
    # 0x6000-0x7FFF 8kb PRG RAM
    # 0x8000-0xBFFF 16kb PRG ROM bank 0 (read only)
    # 0xC000-0xFFFF 16kb PRG ROM bank 1 (read only)
    M1_PRG_RAM_START = 0x6000
    M1_PRG_ROM_BANK0_START = 0x8000
    M1_PRG_ROM_BANK1_START = 0xC000

    # main memory map (write)
    # 0x8000-0xFFFF write only: write a bit to shift register; fifth write commits shift -> internal register, depending
    #                           on address of that fifth write only.
    # Fifth-write addresses:
    # 0x8000-0x9FFF control register
    # 0xA000-0xBFFF CHR bank 0 register
    # 0xC000-0xDFFF CHR bank 1 register
    # 0xE000-0xFFFF PRG bank register
    M1_CTRL_REG_START = 0x8000
    M1_CHR_REG_0_START = 0xA000
    M1_CHR_REG_1_START = 0xC000
    M1_PRG_REG_START = 0xE000

    # vram memory map
    # 0x0000-0x0FFF 4kb CHR ROM bank 0
    # 0x1000-0x1FFF 4kb CHR ROM bank 1
    M1_CHR_ROM_BANK0_START = 0x0000
    M1_CHR_ROM_BANK1_START = 0x1000


cdef class NESCart1(NESCart):
    cdef int num_prg_banks, num_chr_banks, num_prg_ram_banks   # number of available memory banks in this cart
    cdef bint chr_mem_writeable                 # whether or not CHR memory is writeable (i.e. RAM)
    cdef unsigned char ctrl                     # control register
    cdef unsigned char chr_bank[2]              # active chr rom banks
    cdef unsigned char prg_bank, prg_ram_bank   # currently active prg rom/ram bank
    cdef unsigned char shift                    # internal shift register
    cdef int shift_ctr                          # which shift register bit we are currently on

    cdef unsigned char banked_prg_rom[M1_MAX_PRG_BANKS][M1_PRG_ROM_BANK_SIZE]
    cdef unsigned char banked_chr_rom[M1_MAX_CHR_BANKS][M1_CHR_ROM_BANK_SIZE]
    cdef unsigned char ram[M1_MAX_PRG_RAM_BANKS][M1_PRG_RAM_BANK_SIZE]

    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)
    cdef unsigned char read_ppu(self, int address)
    cdef void write_ppu(self, int address, unsigned char value)

    cdef void _write_shift(self, int address, unsigned char value)
    cdef void _set_nametable_mirror_pattern(self)
    cdef int _get_chr_bank(self, int address)


### Mapper 2 (aka UNROM, UOROM) ########################################################################################
# ref: https://wiki.nesdev.com/w/index.php/UxROM

cdef enum:
    # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
    M2_MAX_PRG_BANKS = 16                      # Max number of banks; up to 8 in UNROM, up to 16 in UOROM
    M2_PRG_ROM_BANK_SIZE = 16 * BYTES_PER_KB   # Each bank is 16kB of program ROM

    # memory map
    # 0x8000-0xBFFF:  banked prg_rom
    # 0xC000-0xFFFF:  a single 16kB prg_rom bank fixed to the last bank
    M2_FIXED_PRG_ROM_START = 0xC000


cdef class NESCart2(NESCart0):
    """
    NES Cartridge Type 2 (UNROM / UOROM)
    Much like Mapper 0 above, but has switchable banks of 16kB prg_rom
    """
    # Cython does not let us redefine attributes, so there are a few choices here:
    #   1) use a pointer and allocate memory at instantiation
    #   2) redefine the whole class without inheriting from NESCart0
    #   3) use a different name for prg_rom and waste the 32kB of memory in prg_rom of NESCart0
    # I opt for the third one here because the read and write functions are being rewritten anyway, and this way dynamic
    # memory allocation can be avoided, making the code simpler and safer.  And 32kB is not too much memory to waste.
    cdef unsigned char banked_prg_rom[M2_MAX_PRG_BANKS][M2_PRG_ROM_BANK_SIZE]
    cdef unsigned char prg_bank, num_prg_banks
    cdef bint emulate_bus_conflicts   # whether or not to emulate bus conflicts

    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)


### Mapper 4 (aka MMC3) ########################################################################################
# ref: https://wiki.nesdev.com/w/index.php/MMC3

cdef enum:
    M4_PRG_RAM_SIZE = 8 * BYTES_PER_KB
    M4_PRG_ROM_BANK_SIZE = 8 * BYTES_PER_KB
    M4_CHR_ROM_BANK_SIZE = 1 * BYTES_PER_KB

    M4_MAX_PRG_BANKS = 64
    M4_MAX_CHR_BANKS = 256

    # Main memory map
    M4_PRG_RAM_START = 0x6000
    M4_PRG_ROM_START = 0x8000

    # Registers
    # All these register ranges have two purposes, depending on whether it is an even or odd address
    BANK_REG_START = 0x8000              # even -> bank select, odd -> bank data
    MIRROR_PROTECT_REG_START = 0xA000    # even -> mirroring, odd -> ram protect
    IRQ_LATCH_RELOAD_REG_START = 0xC000  # even -> irq latch, odd -> irq reload
    IRQ_ACTIVATE_START = 0xE000          # even -> irq disable, odd -> irq enable


cdef class NESCart4(NESCart):
    cdef InterruptListener interrupt_listener
    cdef int num_prg_banks, num_chr_banks       # number of available memory banks in this cart
    cdef bint chr_mem_writeable                 # whether or not CHR memory is writeable (i.e. RAM)
    cdef unsigned char bank_register[8]         # bank registers (both chr and prg)

    cdef bint chr_a12_inversion, prg_bank_mode  # mode selectors that determine how bank switching works
    cdef unsigned char bank_select              # determines which bank register to update on next write to bank_data
    cdef bint prg_ram_enable, prg_ram_protect   # prg ram protection flags
    cdef bint mirror_pattern_fixed              # whether the mirror pattern is hard-wired or can be adjusted

    cdef unsigned char irq_reload_value         # the value loaded into the irq latch on reload
    cdef bint irq_reload                        # reload irq counter to reload value at next tick
    cdef bint irq_enabled                       # irq enabled or not?
    cdef int irq_counter                        # counter used to count down to IRQ trigger

    cdef unsigned char banked_prg_rom[M4_MAX_PRG_BANKS][M4_PRG_ROM_BANK_SIZE]
    cdef unsigned char banked_chr_rom[M4_MAX_CHR_BANKS][M4_CHR_ROM_BANK_SIZE]
    cdef unsigned char ram[M4_PRG_RAM_SIZE]

    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)
    cdef unsigned int _get_ppu_bank(self, int address)
    cdef unsigned char read_ppu(self, int address)
    cdef void write_ppu(self, int address, unsigned char value)
    cdef void irq_tick(self)



# cython: profile=True, boundscheck=True, nonecheck=False, language_level=3
from .ppu cimport NESPPU
from .apu cimport NESAPU
from .carts cimport NESCart
from .memory cimport NESMappedRAM
from .mos6502 cimport MOS6502

cdef enum:
    OAM_DMA = 1
    DMC_DMA = 2
    DMC_DMA_DURING_OAM_DMA = 3


cdef class InterruptListener:
    cdef bint _nmi, _irq
    cdef int dma_pause, dma_pause_count

    cdef void raise_nmi(self)
    cdef void reset_nmi(self)
    cdef void raise_irq(self)
    cdef void reset_irq(self)
    cdef void reset_dma_pause(self)
    cdef void raise_dma_pause(self, int type)
    cdef int any_active(self)
    cdef int nmi_active(self)
    cdef int irq_active(self)


cdef enum:
    PPU_CYCLES_PER_CPU_CYCLE = 3
    TARGET_AUDIO_BUFFER_SAMPLES = 2400   # increase this if you get frequent audio glitches, decrease if sound is laggy
    AUDIO_CHUNK_SAMPLES = 400         # how many audio samples go over in each chunk, a frame has 800 samples at 48kHz
    MAX_RATE_DELTA = 3000               # maximum deviation of sample rate +/- from target rate when adaptive
    TARGET_FPS = 60                   # system's target framerate (NTSC)

    # sync modes that are available, each with advantages and disadvantages
    SYNC_NONE = 0      # no sync: runs very fast, unplayable, music is choppy
    SYNC_AUDIO = 1     # sync to audio: rate is perfect, can glitch sometimes, screen tearing can be bad
    SYNC_PYGAME = 2    # sync to pygame's clock, adaptive audio: generally reliable, some screen tearing
    SYNC_VSYNC = 3     # sync to external vsync, adaptive audio: requires ~60Hz vsync, no tearing


cdef class NES:

    cdef:
        NESPPU ppu
        NESAPU apu
        NESCart cart
        MOS6502 cpu
        NESMappedRAM memory
        InterruptListener interrupt_listener

        object controller1, controller2
        object screen

        int screen_scale, sync_mode
        bint v_overscan, h_overscan

    cdef int step(self, int log_cpu)
    cpdef void run(self)
    cpdef object run_frame_headless(self, int run_frames=?, object controller1_state=?, object controller2_state=?)



from .mos6502 import MOS6502
from .memory import BigEmptyRAM, NESMappedRAM
from .rom import ROM

from .utils import load_rom_raw

def setup_machine_basic():
    program_start = 0xC000
    load_at = 0xC000 - 16   # don't know why the rom starts here, but it does
    memory = BigEmptyRAM()
    cpu = MOS6502(memory, support_BCD=False)  # don't support BCD on NES
    rom, header = load_rom_raw("./testroms/nestest/nestest.nes")

    # note that the rom is too long and overruns the memory so have to add the reset vector after
    # I don't know why this is (it is a NES rom) or how it should be loaded properly
    memory.ram[load_at:load_at + len(rom)] = rom

    cpu.set_reset_vector(program_start)

    return cpu

def setup_machine_nes():
    rom = ROM("./testroms/nestest/nestest.nes")
    cart = rom.get_cart(prg_start=0xC000)  #NESCart0(rom_data=rom_data, load_rom_at=0xC000-16)
    cart.prg_start_addr = 0xC000
    memory = NESMappedRAM(cart=cart)
    cpu = MOS6502(memory, support_BCD=False)  # don't support BCD on NES
    cpu.set_reset_vector(0xC000)

    return cpu

def nestest(num_instrs, suppress_until=0):
    cpu = setup_machine_nes()

    error_lines = []

    # begin the test
    with open("./testroms/nestest/nestest.log", "rt") as f:

        cpu.reset()

        cpulog = cpu.log_line()
        truelog = f.readline()
        if suppress_until <= 0:
            print(cpulog)
            print(truelog[:-1])

        # compare the crucial last bit of the lines, excluding the PPU location:
        if cpulog[48:74] != truelog[48:74] or cpulog[87:] != truelog[87:-1]:
            error_lines.append(0)
            print(" *************** ERROR ")

        for i in range(1, num_instrs):

            if i >= suppress_until:
                print()
                print(i)
            cpu.run_next_instr()

            cpulog = cpu.log_line()
            truelog = f.readline()

            if len(truelog) == 0:
                print("FINISHED ALL TESTS - CONGRATULATIONS")
                break

            if i >= suppress_until:
                print(cpulog[:46] + cpulog[48:])   # cut out some whitespace to make it fit on a line
                print(truelog[:46] + truelog[48:-1])
                # memory.print(0x00, 256)
                # print("NV-BDIZC")
                # print("{:08b}".format(cpu._status_to_byte()))

            # compare the crucial last bit of the lines, excluding (for now) the PPU states:
            if cpulog[48:74] != truelog[48:74] or cpulog[87:] != truelog[87:-1]:
                error_lines.append(i)
                if i >= suppress_until:
                    print("\033[93m***** ERROR *************** ERROR *************** ERROR *************** ERROR ***************\033[0m")

    print()
    print("ERRORS AT:")
    print(error_lines)




from .system cimport InterruptListener
from .memory cimport NESMappedRAM

cdef enum:
    # sound synthesis constants
    SAMPLE_RATE = 48000     # 48kHz sample rate
    SAMPLE_SCALE = 65536    # 16 bit samples
    SAMPLE_OFFSET = 32768
    CPU_FREQ_HZ = 1789773   # https://wiki.nesdev.com/w/index.php/Cycle_reference_chart#Clock_rates
    MAX_CPU_CYCLES_PER_LOOP = 24  # if the cpu has done more than this many cycles, complete them in loops

    # control registers
    STATUS = 0x4015
    FRAME_COUNTER = 0x4017

    # counter modes
    FOUR_STEP = 0
    FIVE_STEP = 1

    # bits in the status register during read
    BIT_DMC_INTERRUPT = 7
    BIT_FRAME_INTERRUPT = 6
    BIT_DMC_ACTIVE = 4
    BIT_LENGTH_NOISE = 3
    BIT_LENGTH_TRIANGLE = 2
    BIT_LENGTH_PULSE2 = 1
    BIT_LENGTH_PULSE1 = 0

    # Bits in the status register during write
    BIT_ENABLE_DMC = 4
    BIT_ENABLE_NOISE = 3
    BIT_ENABLE_TRIANGLE = 2
    BIT_ENABLE_PULSE2 = 1
    BIT_ENABLE_PULSE1 = 0

    # Bits in the frame counter register
    BIT_MODE = 7
    BIT_IRQ_INHIBIT = 6

    # envelopes
    NUM_ENVELOPES = 3
    PULSE0 = 0  # } use 0 and 1 for the pulse to match with pulse_ix
    PULSE1 = 1  # }
    NOISE = 2

    # buffer length of the APU's output buffer; must be a power of 2
    APU_BUFFER_LENGTH = 65536
    CHUNK_SIZE = 10000


cdef class APUEnvelope:
    cdef:
        bint start_flag, loop_flag
        unsigned int decay_level, divider, volume

    cdef void update(self)
    cdef void restart(self)


cdef class APUUnit:
    cdef:
        bint enable, ctr_halt
        int length_ctr
        short output[SAMPLE_RATE]

    cdef void update_length_ctr(self)
    cdef void set_enable(self, bint value)
    cdef void set_length_ctr(self, int value)


cdef class APUTriangle(APUUnit):
    cdef:
        bint linear_reload_flag
        int linear_reload_value, period, linear_ctr
        double phase

    cdef void write_register(self, int address, unsigned char value)
    cdef void quarter_frame(self)
    cdef void half_frame(self)
    cdef int generate_sample(self)


cdef class APUPulse(APUUnit):
    cdef:
        bint constant_volume, is_unit_1
        int period, adjusted_period, duty
        double phase
        APUEnvelope env
        bint sweep_enable, sweep_negate, sweep_reload
        int sweep_period, sweep_shift, sweep_divider

        int duty_waveform[4][8]  # duty cycle sequences

    cdef void write_register(self, int address, unsigned char value)
    cdef void sweep_update(self)
    cdef void quarter_frame(self)
    cdef void half_frame(self)
    cdef int generate_sample(self)


cdef class APUNoise(APUUnit):
    cdef:
        bint constant_volume, mode
        int period, feedback, timer
        #double shift_ctr
        APUEnvelope env

        unsigned int timer_table[16]    # noise timer periods

    cdef void write_register(self, int address, unsigned char value)
    cdef void update_cycles(self, int cycles)
    cdef void quarter_frame(self)
    cdef void half_frame(self)
    cdef int generate_sample(self)


cdef class APUDMC(APUUnit):
    """
    The DMC unit is pretty different to the other APU units
    """
    cdef:
        NESMappedRAM memory
        InterruptListener interrupt_listener

        bint irq_enable, loop_flag, silence, interrupt_flag
        unsigned int sample_address, sample_length, address, bytes_remaining
        unsigned int rate, timer
        int output_level, bits_remaining
        unsigned char sample

        unsigned int rate_table[16]    # sample consumption rates for the dmc

    cdef void write_register(self, int address, unsigned char value)
    cdef void update_cycles(self, int cycles)
    cdef void read_advance(self)
    cdef int generate_sample(self)


cdef class NESAPU:
    """
    References:
        [1] https://wiki.nesdev.com/w/index.php/APU#Registers
        [2] https://wiki.nesdev.com/w/index.php/APU_Frame_Counter
    """
    cdef:
        #### master volume
        double master_volume

        #### apu state variables
        int cycles,rate  # cycle within the current frame (counted in CPU cycles NOT APU cycles as specified in [2]
        int frame_segment  # which segment of the frame the apu is in
        int _reset_timer_in  # after this number of cycles, reset the timer; ignored if < 0
        double samples_per_cycle  # number of output samples to generate per output cycle (will be <1)
        double samples_required  # number of samples currently required, once this gets over 1, generate a sample
        unsigned long long _buffer_start, _buffer_end  # start and end index of the sample ring buffer

        #### system interrupt listener
        InterruptListener interrupt_listener

        #### buffers for up to 1s of data for each of the waveform generators
        short output[APU_BUFFER_LENGTH]   # final output from the mixer; power of two sized to make ring buffer easier to implement
        short buffer[CHUNK_SIZE]

        #### status register
        bint mode, irq_inhibit, frame_interrupt_flag

        #### Sound units
        APUTriangle triangle
        APUPulse pulse1, pulse2
        APUNoise noise
        APUDMC dmc

        #### Frame counters
        int frame_counter

    ##########################################################################

    # register control functions
    cdef unsigned char read_register(self, int address)
    cdef void write_register(self, int address, unsigned char value)
    cdef void _set_status(self, unsigned char value)

    # synchronous update functions
    cdef int run_cycles(self, int cpu_cycles)
    cdef void quarter_frame_tick(self)
    cdef void half_frame_tick(self)

    # mixer
    cdef int mix(self, int tri, int p1, int p2, int noise, int dmc)

    # output
    cdef void generate_sample(self)
    cpdef short[:] get_sound(self, int samples)
    cpdef void set_volume(self, float volume)

    cpdef int buffer_remaining(self)

    cpdef void set_rate(self, int rate)
    cpdef int get_rate(self)




import logging

# would like to make this not depend on pygame
try:
    import pygame
    import pygame.freetype
    has_pygame = True
except ImportError:
    has_pygame = False

try:
    from OpenGL.GL import *
    has_opengl = True
except ImportError:
    has_opengl = False

class ScreenBase:
    """
    Base class for screens.  Not library specific.
    """
    WIDTH_PX = 256
    HEIGHT_PX = 240
    VISIBLE_HEIGHT_PX = 224
    VISIBLE_WIDTH_PX = 240

    def __init__(self, ppu, scale=3, vertical_overscan=False, horizontal_overscan=False):
        self.ppu = ppu
        self.width = self.WIDTH_PX if horizontal_overscan else self.VISIBLE_WIDTH_PX
        self.height = self.HEIGHT_PX if vertical_overscan else self.VISIBLE_HEIGHT_PX
        self.scale = scale
        self.vertical_overscan = vertical_overscan
        self.horizontal_overscan = horizontal_overscan

        self._text_buffer = []

    def show(self):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()

    def add_text(self, text, position, color, ttl=1):
        self._text_buffer.append((text, position, color, ttl))

    def update_text(self):
        self._text_buffer = [(txt, pos, col, ttl - 1) for (txt, pos, col, ttl) in self._text_buffer if ttl > 1]


class Screen(ScreenBase):
    """
    PyGame based screen.
    Keep PyGame-specific stuff in here (don't want PyGame specific stuff all over the rest of the code)
    """
    def __init__(self, ppu, scale=3, vsync=False, vertical_overscan=False, horizontal_overscan=False, nametable_panel=False, py_compatibility_mode=False):
        super().__init__(ppu, scale, vertical_overscan, horizontal_overscan)

        self.nametable_panel = nametable_panel
        self.py_compatibility_mode = py_compatibility_mode

        if nametable_panel:
            # if including the nametable panel, make a buffer for it
            self.nt_buffer_surf = pygame.Surface((32 * 8 * 2, 30 * 8 * 2))

        # screens and buffers
        self.buffer_surf = pygame.Surface((self.width, self.height))

        # todo: this introduces a dependency on NumPy, so we might want to refactor to remove that
        self.buffer_sa = pygame.surfarray.pixels2d(self.buffer_surf)

        if nametable_panel:
            # if including the nametable panel, make the window large enough to show it
            scr_width = self.width * self.scale + 32 * 8 * 2
            scr_height = max(self.height * self.scale, 30 * 8 * 2)
        else:
            scr_width = self.width * self.scale
            scr_height = self.height * self.scale

        self.screen = pygame.display.set_mode((scr_width, scr_height), vsync=vsync)

        # font for writing to HUD
        pygame.freetype.init()
        self.font = pygame.freetype.SysFont(pygame.font.get_default_font(), 12 * self.scale)

    def write_at(self, x, y, color):
        # only used in py_compatibility_mode
        self.buffer_surf.set_at((x, y), color)

    def add_text(self, text, position, color, ttl=1):
        self._text_buffer.append((text, (position[0], position[1]), color, ttl))

    def _render_text(self, surf):
        for (text, position, color, _) in self._text_buffer:
            self.font.render_to(surf, (position[0] * self.scale, position[1] * self.scale), text, color)

    def show(self):
        if not self.py_compatibility_mode:
            # the old version of the ppu wrote directly to the screen via write_at
            self.ppu.copy_screen_buffer_to(self.buffer_sa, self.vertical_overscan, self.horizontal_overscan)
        if self.nametable_panel:
            self.ppu.debug_render_nametables(pygame.surfarray.pixels2d(self.nt_buffer_surf))
            scaled = pygame.transform.scale(self.buffer_surf, (self.width * self.scale, self.height * self.scale))
            self.screen.blit(scaled, (0,0))
            self.screen.blit(self.nt_buffer_surf, (self.width * self.scale, 0))
        else:
            pygame.transform.scale(self.buffer_surf, (self.width * self.scale, self.height * self.scale), self.screen)
        self._render_text(self.screen)
        pygame.display.flip()
        self.update_text()

    def clear(self, color=(0, 0, 0)):
        self.buffer_surf.fill(color)


class ScreenGL(ScreenBase):
    """
    PyGame / OpenGL based screen.
    Keep *all* OpenGL-specific stuff in here
    Keep PyGame-specific stuff in here (don't want PyGame specific stuff all over the rest of the code)
    """

    def __init__(self, ppu, scale=3, vsync=False, vertical_overscan=False, horizontal_overscan=False):
        super().__init__(ppu, scale, vertical_overscan, horizontal_overscan)

        # screens and buffers
        self.buffer_surf = pygame.Surface((self.width, self.height))
        self.buffer_sa = pygame.surfarray.pixels2d(self.buffer_surf)
        self.screen = pygame.display.set_mode((self.width * scale, self.height * scale),
                                              flags = pygame.DOUBLEBUF | pygame.OPENGL,
                                              vsync=vsync
                                              )

        self.arr = bytearray([0] * (self.width * self.height * 3))
        self.gltex = None   # the texture that we will use for the screen
        self.opengl_init()

        # font for writing to HUD
        pygame.freetype.init()
        self.font = pygame.freetype.SysFont(pygame.font.get_default_font(), 12)
        self._text_buffer = []

    def opengl_init(self):
        """
        Set up the OpenGL boilerplate to get going
        """
        self.gltex = glGenTextures(1)
        glViewport(0, 0, self.width * self.scale, self.height * self.scale)
        glDepthRange(0, 1)
        glMatrixMode(GL_PROJECTION)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        glShadeModel(GL_SMOOTH)
        glClearColor(0.0, 0.0, 0.0, 0.0)
        glClearDepth(1.0)
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glDepthFunc(GL_LEQUAL)
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        glEnable(GL_BLEND)

    def show_gl(self):
        """
        Show the screen by copying the buffer to an OpenGL texture and displaying a rectangle with that texture
        :return:
        """
        # prepare to render the texture-mapped rectangle
        glClear(GL_COLOR_BUFFER_BIT)
        glLoadIdentity()
        glDisable(GL_LIGHTING)
        glEnable(GL_TEXTURE_2D)

        # draw texture openGL Texture
        self.surface_to_texture(self.buffer_surf)
        glBindTexture(GL_TEXTURE_2D, self.gltex)
        glBegin(GL_QUADS)
        glTexCoord2f(0, 0); glVertex2f(-1, 1)
        glTexCoord2f(0, 1); glVertex2f(-1, -1)
        glTexCoord2f(1, 1); glVertex2f(1, -1)
        glTexCoord2f(1, 0); glVertex2f(1, 1)
        glEnd()

    def surface_to_texture(self, pygame_surface):
        """
        Copy a PyGame surface to an OpenGL texture.  This came from a StackOverflow answer that I sadly can't find now.
        There is probably a faster way to do this, but this works for now
        """
        rgb_surface = pygame.image.tostring(pygame_surface, 'RGB')
        glBindTexture(GL_TEXTURE_2D, self.gltex)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
        surface_rect = pygame_surface.get_rect()
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, surface_rect.width, surface_rect.height, 0, GL_RGB, GL_UNSIGNED_BYTE,
                     rgb_surface)
        glGenerateMipmap(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, 0)

    def _render_text(self, surf):
        for (text, position, color, _) in self._text_buffer:
            self.font.render_to(surf, position, text, color)

    def show(self):
        self.ppu.copy_screen_buffer_to(self.buffer_sa)
        self._render_text(self.buffer_surf)
        self.show_gl()
        pygame.display.flip()
        self.update_text()

    def clear(self, color=(0, 0, 0)):
        self.buffer_surf.fill(color)


class ControllerBase:
    """
    NES Controller (no Pygame code in here)

    References:
        [1] https://wiki.nesdev.com/w/index.php/Standard_controller
    """
    # code for each button
    # this is not just an enum, this is the bit position that they are fed out of the controller
    A = 0
    B = 1
    SELECT = 2
    START = 3
    UP = 4
    DOWN = 5
    LEFT = 6
    RIGHT = 7

    NUM_BUTTONS = 8
    NAMES = ['A', 'B', 'select', 'start', 'up', 'down', 'left', 'right']

    def __init__(self, active=True):
        self.is_pressed = [0] * 8   # array to store key status
        self._current_bit = 0
        self.strobe = False
        self.active = active  # allows the gamepad to be turned off (acting as if it were disconnected)

    def update(self):
        pass

    def set_state(self, state):
        """
        Sets the controller state from a length-8 array of boolean values.  Set this way to support any type of array as
        input state.
        """
        self.is_pressed[self.A] = state[self.A]
        self.is_pressed[self.B] = state[self.B]
        self.is_pressed[self.SELECT] = state[self.SELECT]
        self.is_pressed[self.START] = state[self.START]
        self.is_pressed[self.UP] = state[self.UP]
        self.is_pressed[self.DOWN] = state[self.DOWN]
        self.is_pressed[self.LEFT] = state[self.LEFT]
        self.is_pressed[self.RIGHT] = state[self.RIGHT]

    def set_strobe(self, value):
        """
        Set the strobe bit to the given value
        """
        # we don't need to do much with the strobe, just reset the status bit if strobe is high so that we start
        # out at bit 0.  If strobe is low, do nothing; then we can read out the data from the ouptut port.
        self.strobe = value
        if value:
            self._current_bit = 0

    def read_bit(self):
        """
        Read a bit from the gamepad.  Buttons are read through a series of serial reads.
        "The first 8 reads will indicate which buttons or directions are pressed (1 if pressed, 0 if not pressed).
        All subsequent reads will return 1 on official Nintendo brand controllers but may return 0 on third party
        controllers" [1]
        :return:
        """
        if not self.active:
            return 0

        # todo: it seems like the following behviour should be correct, but it doesn't work, whereas removing this does
        #if self.strobe:
        #    self._current_bit = 0

        v = self.is_pressed[self._current_bit] if self._current_bit < self.NUM_BUTTONS else 1
        #logging.log(logging.DEBUG, "Controller bit {} is {}".format(self._current_bit, v), extra={"source": "cntrlr"})
        self._current_bit = min((self._current_bit + 1), self.NUM_BUTTONS) # don't want this to overflow (very unlikely)
        return v


class KeyboardController(ControllerBase):
    """
    PyGame keyboard-based controller
    Keep PyGame specific stuff in here
    """
    def __init__(self, active=True, key_map=None):
        super().__init__(active=active)
        self.key_map = key_map if key_map is not None else self._default_key_map()

    def _default_key_map(self):
        """
        Make a default key map; done this way to avoid using pygame in a class-level constant, which causes an error
        if pygame is not available.
        """
        return {
            pygame.K_w: ControllerBase.UP,
            pygame.K_a: ControllerBase.LEFT,
            pygame.K_s: ControllerBase.DOWN,
            pygame.K_d: ControllerBase.RIGHT,
            pygame.K_g: ControllerBase.SELECT,
            pygame.K_h: ControllerBase.START,
            pygame.K_l: ControllerBase.B,
            pygame.K_p: ControllerBase.A,
        }

    def update(self):
        """
        This should get called once every game loop and updates the internal status of the gamepad
        Read the keyboard and put the status of the keys into the key_pressed array.
        """
        keys = pygame.key.get_pressed()
        for k, v in self.key_map.items():
            self.is_pressed[v] = keys[k]



# import pyximport; pyximport.install()

from .mos6502 cimport MOS6502
from .memory cimport NESMappedRAM
from .ppu cimport NESPPU
from .apu cimport NESAPU, SAMPLE_RATE

from nes.rom import ROM
from nes.peripherals import Screen, ScreenGL, KeyboardController, ControllerBase
from nes.utils import load_palette

import logging
import time
import warnings

# unfortunately, pygame's audio handling does not do what we need with gapless playback of short samples
# we therefore use pyaudio, but it would be nice to make it optional
try:
    import pyaudio
    has_audio = True
except ImportError:
    has_audio = False

# would like to make this not depend on pygame
try:
    import pygame
    has_pygame = True
except ImportError:
    has_pygame = False

# would like to know if PyOpenGL is available to avoid using OpenGL screen if not
try:
    import OpenGL.GL
    has_opengl = True
except ImportError:
    has_opengl = False

# only use numpy for return of frame during headless operation, so don't depend on this in general
try:
    import numpy as np
    has_numpy = True
except ImportError:
    has_numpy = False



# custom log levels to control the amount of logging (these are mostly unused because logging is mostly disabled)
LOG_MEMORY = 5
LOG_PPU = 6
LOG_CPU = 7

cdef class InterruptListener:
    def __init__(self):
        self._nmi = False
        self._irq = False   # the actual IRQ line on the 6502 rests high and is low-triggered
        self.dma_pause = False
        self.dma_pause_count = 0

    cdef void raise_nmi(self):
        self._nmi = True

    cdef void reset_nmi(self):
        self._nmi = False

    cdef void raise_irq(self):
        self._irq = True

    cdef void reset_irq(self):
        self._irq = False

    cdef void reset_dma_pause(self):
        self.dma_pause = 0
        self.dma_pause_count = 0

    cdef void raise_dma_pause(self, int type):
        self.dma_pause = type
        self.dma_pause_count += 1  # can (and this will be rare) get multiple DMC DMA pauses during OAM DMA

    cdef int any_active(self):
        return self._nmi or self._irq or self.dma_pause

    cdef int nmi_active(self):
        return self._nmi

    cdef int irq_active(self):
        return self._irq


cdef class NES:
    """
    The NES system itself, combining all of the parts, and the interlinks
    """
    OSD_Y = 3
    OSD_FPS_X = 5
    OSD_VOL_X = 110
    OSD_MUTE_X = 200
    OSD_NOTE_X = 100

    OSD_TEXT_COLOR = (0, 255, 0)
    OSD_NOTE_COLOR = (255, 128, 0)
    OSD_WARN_COLOR = (255, 0, 0)

    STATS_CALC_PERIOD_S = 1.0        # how often to calculate the adaptive audio rate and fps counter

    def __init__(self,
                 rom_file,                  # the rom file to load
                 screen_scale=3,            # factor by which to scale the screen
                 log_file=None,             # file to log to (logging is largely turned off by default)
                 log_level=None,            # level of logging (logging is largely turned off by default)
                 opengl=False,              # use opengl for screen rendering
                 sync_mode=SYNC_AUDIO,      # audio / video sync mode
                 verbose=True,              # whether to print out cartridge info at startup
                 show_nametables=False,     # shows the nametables alongside the main screen (for debug, not opengl)
                 vertical_overscan=False,   # show the top and bottom 8 pixels (not usually visible on CRT TVs)
                 horizontal_overscan=False, # show the left and right 8 pixels (often not visible on CRT TVs)
                 palette_file=None,         # supply a palette file to use; None gives default
                 headless=False,            # runs the nes in headless mode without the pygame screen being started
                 ):
        """
        Build a NES and cartridge from the bits and pieces we have lying around plus some rom data!  Also do some things
        like set up logging, make sure the options are compatible (and warn if not) etc.
        """
        if sync_mode == SYNC_AUDIO and not has_audio:
            self.sync_mode = SYNC_PYGAME
            warnings.warn("Selected sync_mode is SYNC_AUDIO, but audio is not available, probably because pyaudio is "
                          "not available on the system")
        else:
            self.sync_mode = sync_mode

        if opengl and not has_opengl:
            warnings.warn("OpenGL is not available, possibly because PyOpenGL is not installed (try 'pip install "
                          "PyOpenGL' or start with option opengl=False).  Starting with PyGame (SDL) based screen.")

        if opengl and show_nametables:
            warnings.warn("Nametable display not supported in OpenGL mode.  Use opengl=False to display.")

        # record overscan preference
        self.h_overscan = horizontal_overscan
        self.v_overscan = vertical_overscan

        # set up the logger
        self.init_logging(log_file, log_level)

        # the interrupt listener here is not a NES hardware device, it is an interrupt handler that is used to pass
        # interrupts between the PPU, APU and sometimes even cartridge and CPU in this emulator
        self.interrupt_listener = InterruptListener()

        rom = ROM(rom_file, verbose=verbose)
        # the cartridge is a piece of hardware (unlike the ROM, which is just data) and must come first because it
        # supplies bits of hardware (memory in most cases, both ROM and RAM) all over the system, so it
        # affects the actual hardware configuration of the system.
        self.cart = rom.get_cart(self.interrupt_listener)

        # game controllers have no dependencies
        if has_pygame:
            self.controller1 = KeyboardController()
        else:
            self.controller1 = ControllerBase(active=True)
        self.controller2 = ControllerBase(active=True)   # connect a second gamepad, but make it inactive for now

        # load the requested palette
        if palette_file is not None:
            palette = load_palette(palette_file)
        else:
            palette = None

        # set up the APU and the PPU
        self.ppu = NESPPU(cart=self.cart, interrupt_listener=self.interrupt_listener, palette=palette)
        self.apu = NESAPU(interrupt_listener=self.interrupt_listener)

        # screen needs to have the PPU
        if has_pygame and not headless:
            if opengl and has_opengl:
                self.screen = ScreenGL(ppu=self.ppu,
                                       scale=screen_scale,
                                       vsync=True if sync_mode == SYNC_VSYNC else False,
                                       vertical_overscan=vertical_overscan,
                                       horizontal_overscan=horizontal_overscan
                                      )
            else:
                self.screen = Screen(ppu=self.ppu,
                                     scale=screen_scale,
                                     vsync=True if sync_mode == SYNC_VSYNC else False,
                                     nametable_panel=show_nametables,
                                     vertical_overscan=vertical_overscan,
                                     horizontal_overscan=horizontal_overscan
                                    )
        else:
            self.screen=None

        self.screen_scale = screen_scale

        # due to memory mapping, lots of things are connected to the main memory
        self.memory = NESMappedRAM(ppu=self.ppu,
                                   apu=self.apu,
                                   cart=self.cart,
                                   controller1=self.controller1,
                                   controller2=self.controller2,
                                   interrupt_listener=self.interrupt_listener
                                   )

        # one nasty wrinkle here is that we have to give the apu's dmc channel access to the memory:
        self.apu.dmc.memory = self.memory

        # only the memory is connected to the cpu, all access to other devices is done through memory mapping
        self.cpu = MOS6502(memory=self.memory,
                           undocumented_support_level=2,  # a few NES games make use of some more common undocumented instructions
                           stack_underflow_causes_exception=False
                           )

        # Let's get started!  Reset the cpu so we are ready to go...
        self.cpu.reset()

    def init_logging(self, log_file, log_level):
        """
        Initialize the logging; set the log file and the logging level (LOG_MEMORY, LOG_PPU, LOG_CPU are all below
        logging.DEBUG)
        """
        if log_file is None or log_level is None:
            logging.disable(level=logging.CRITICAL)
            return

        logging.addLevelName(LOG_MEMORY, "MEMORY")  # set a low level for memory logging because it is so intense
        logging.addLevelName(LOG_PPU, "PPU")
        logging.addLevelName(LOG_CPU, "CPU")
        # set up logging to the format as we required
        logging.basicConfig(filename=log_file,
                            level=logging.NOTSET,
                            format='%(asctime)-15s %(source)-5s %(message)s',
                            filemode='w',
                            )
        logging.root.setLevel(log_level)

    cdef int step(self, int log_cpu):
        """
        The heartbeat of the system.  Run one instruction on the CPU and the corresponding amount of cycles on the
        PPU (three per CPU cycle, at least on NTSC systems).
        """
        cdef int cpu_cycles=0
        cdef bint vblank_started=False

        if self.interrupt_listener.any_active():
            if self.interrupt_listener.nmi_active():
                cpu_cycles = self.cpu.trigger_nmi()  # raises an NMI on the CPU, but this does take some CPU cycles
                # should we do this here or leave this up to the triggerer?  It's really up to the triggerer to put the NMI
                # line into its "off" position, but if the line remains in the same state the CPU will not trigger another
                # NMI anyway.  Electronics lend themselves to edge-triggered events, whereas software is more naturally
                # pulse triggered in some ways (a thing happens then returns).
                self.interrupt_listener.reset_nmi()
                if log_cpu:
                    logging.log(logging.INFO, "NMI Triggered", extra={"source": "Interrupt"})
            elif self.interrupt_listener.irq_active():
                # note, cpu_cycles can be zero here if the cpu is set to ignore irq
                cpu_cycles = self.cpu.trigger_irq()
                if cpu_cycles > 0:
                    # the cpu reacted to the interrupt, so clear it here so that it won't re-trigger the CPU;
                    # if the cpu does not act on this interrupt, it will remain here (unless cleared by the caller) and
                    # will try to trigger again on the next cycle
                    self.interrupt_listener.reset_irq()
                    if log_cpu:
                        logging.log(logging.INFO, "IRQ Triggered", extra={"source": "Interrupt"})
            elif self.interrupt_listener.dma_pause:
                # https://wiki.nesdev.com/w/index.php/PPU_OAM#DMA
                cpu_cycles = self.cpu.dma_pause(self.interrupt_listener.dma_pause,
                                                self.interrupt_listener.dma_pause_count
                                                )
                if log_cpu:
                    logging.log(logging.INFO,
                                "OAM DMA (type: {}, count: {})".format(self.interrupt_listener.dma_pause,
                                                                       self.interrupt_listener.dma_pause_count),
                                extra={"source": "Interrupt"}
                                )

                self.interrupt_listener.reset_dma_pause()
                had_dma_pause = True
        if cpu_cycles == 0:
            cpu_cycles = self.cpu.run_next_instr()
            if log_cpu:
                logging.log(logging.INFO, self.cpu.log_line(self.ppu.line, self.ppu.pixel) , extra={"source": "CPU"})

        vblank_started = self.ppu.run_cycles(cpu_cycles * PPU_CYCLES_PER_CPU_CYCLE)
        self.apu.run_cycles(cpu_cycles)
        if had_dma_pause and self.interrupt_listener.dma_pause:
            # any apu dmc dma pause that occurred during an oam dma pause should average 2 rather than 4 cycles
            self.interrupt_listener.dma_pause = DMC_DMA_DURING_OAM_DMA
        return vblank_started

    cpdef void run(self):
        """
        Run the NES indefinitely (or until some quit signal); this will only exit on quit.
        There is some PyGame specific stuff in here in order to handle frame timing and checking for exits
        """
        cdef int vblank_started=False
        cdef float volume = 0.5
        cdef double fps, t_start=0., dt=-0.
        cdef bint show_hud=True, log_cpu=False, mute=False, audio_drop=False
        cdef int frame=0, frame_start=0, cpu_cycles=0, adaptive_rate=0, buffer_surplus=0
        cdef bint audio=has_audio

        if not has_pygame:
            raise RuntimeError("Cannot run() without pygame; only headless operation is supported.")

        if audio:
            p = pyaudio.PyAudio()
        else:
            warnings.warn("Audio is unavailable, probably because pyaudio is not installed.")

        pygame.init()
        clock = pygame.time.Clock()

        # this has to come after pygame init for some reason, or pygame won't start :(
        if audio:
            player = p.open(format=pyaudio.paInt16,
                            channels=1,
                            rate=SAMPLE_RATE,
                            output=True,
                            frames_per_buffer=AUDIO_CHUNK_SAMPLES,  # 400 is a half-frame at 60Hz, 48kHz sound
                            stream_callback=self.apu.pyaudio_callback,
                            )
            player.start_stream()

        t_start = time.time()

        while True:
            vblank_started=False
            while not vblank_started:
                vblank_started = self.step(log_cpu)

            # update the controllers once per frame
            self.controller1.update()
            self.controller2.update()

            cpu_cycles = self.cpu.cycles_since_reset

            if time.time() - t_start > self.STATS_CALC_PERIOD_S:
                # calcualte the fps and adaptive audio rate (only used when non-audio sync is in use)
                dt = time.time() - t_start
                fps = (frame - frame_start) * 1.0 / dt
                buffer_surplus = self.apu.buffer_remaining() - TARGET_AUDIO_BUFFER_SAMPLES

                # adjust the rate based on the buffer change
                adaptive_rate = int(SAMPLE_RATE - 0.5 * buffer_surplus / dt)
                adaptive_rate = max(SAMPLE_RATE - MAX_RATE_DELTA, min(adaptive_rate, SAMPLE_RATE + MAX_RATE_DELTA))
                t_start = time.time()
                frame_start = frame
            if show_hud:
                # display information on the HUD (turn on/off with '1' key)
                self.screen.add_text("{:.0f} fps, {}Hz, {} samples".format(fps, self.apu.get_rate(), self.apu.buffer_remaining()),
                                     (self.OSD_FPS_X, self.OSD_Y),
                                     self.OSD_TEXT_COLOR if fps > TARGET_FPS - 3 else self.OSD_WARN_COLOR)
                if log_cpu:
                    self.screen.add_text("logging cpu", (self.OSD_NOTE_X, self.OSD_Y), self.OSD_NOTE_COLOR)
                if mute:
                    self.screen.add_text("MUTE", (self.OSD_MUTE_X, self.OSD_Y), self.OSD_TEXT_COLOR)

            # Check for an exit
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    if audio:
                        player.stop_stream()
                        player.close()
                        p.terminate()
                    return
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_1:
                        show_hud = not show_hud
                    if event.key == pygame.K_0:
                        if not mute:
                            self.apu.set_volume(0)
                            mute = True
                        else:
                            self.apu.set_volume(volume)
                            mute = False
                    if event.key == pygame.K_MINUS:
                        volume = max(0, volume - 0.1)
                        self.apu.set_volume(volume)
                        self.screen.add_text("volume: " + "|" * int(10 * volume),
                                             (self.OSD_VOL_X, self.OSD_Y),
                                             self.OSD_TEXT_COLOR, ttl=30)
                        mute=False
                    if event.key == pygame.K_EQUALS:
                        volume = min(1, volume + 0.1)
                        self.apu.set_volume(volume)
                        self.screen.add_text("volume: " + "|" * int(10 * volume),
                                             (self.OSD_VOL_X, self.OSD_Y),
                                             self.OSD_TEXT_COLOR, ttl=30)
                        mute=False
                    if event.key == pygame.K_2:
                        log_cpu = not log_cpu


            # show the display (if using SYNC_VSYNC mode, this should provide a sync, which must be at 60Hz)
            self.screen.show()

            if self.sync_mode == SYNC_AUDIO:
                # wait for the audio buffer to empty, but only if the audio is playing
                while self.apu.buffer_remaining() > TARGET_AUDIO_BUFFER_SAMPLES and audio and player.is_active():
                    clock.tick(500)  # wait for about 2ms (~= 96 samples)
            elif self.sync_mode == SYNC_VSYNC or self.sync_mode == SYNC_PYGAME:
                # here we rely on an external sync source, but allow the audio to adapt to it
                if frame > 2 * TARGET_FPS:  # wait a bit before doing this since startup can be slow
                    self.apu.set_rate(adaptive_rate)
            else:
                # no sync at all, go as fast as we can!
                pass

            if self.sync_mode == SYNC_PYGAME:
                # if we are using pygame sync, we have to supply our own clock tick here
                clock.tick(TARGET_FPS)

            if (audio
                and not player.is_active()
                and self.apu.buffer_remaining() > TARGET_AUDIO_BUFFER_SAMPLES
                and (frame % 10 == 0)
                ):
                # sometimes the audio stream stops (e.g. if it runs out of samples) and has to be restarted or else the
                # sound will stop.  Here try to (re)start the stream if it is not running if there is audio waiting.
                audio_drop=True
                player.stop_stream()
                player.close()
                player = p.open(format=pyaudio.paInt16,
                                channels=1,
                                rate=SAMPLE_RATE,
                                output=True,
                                frames_per_buffer=AUDIO_CHUNK_SAMPLES,  # 400 is a half-frame at 60Hz, 48kHz sound
                                stream_callback=self.apu.pyaudio_callback,
                                )
                player.start_stream()
            else:
                audio_drop = False

            frame += 1

    cpdef object run_frame_headless(self, int run_frames=1, object controller1_state=[False] * 8, object controller2_state=[False] * 8):
        """
        Runs a single frame of the emulator and returns an array containing the screen output.
        :param run_frames: number of frames to run; controller will be in same state for all frames
        :param controller1_state: controller1 state as an array in order A, B, select, start, up, down, left, right
        :param controller2_state: controller2 state as an array in order A, B, select, start, up, down, left, right
        :return:
        """
        cdef int w, h, frame
        cdef bint vblank_started

        h = 240 if self.v_overscan else 224
        w = 256 if self.h_overscan else 240
        buffer = np.zeros((w, h), dtype=np.uint32)
        cdef unsigned int [:, :] buffer_mv = buffer

        # set the controller state (same for all frames)
        self.controller1.set_state(controller1_state)
        self.controller2.set_state(controller2_state)

        frame = 0
        while frame < run_frames:
            vblank_started=False
            while not vblank_started:
                vblank_started = self.step(log_cpu=False)
            frame += 1

        self.ppu.copy_screen_buffer_to(buffer_mv, v_overscan=self.v_overscan, h_overscan=self.h_overscan)

        # converts to an RGB buffer with one channel per color
        buffer_rgb = buffer.view(dtype=np.uint8).reshape((w, h, 4))[:, :, np.array([2, 1, 0])].swapaxes(0, 1)
        return buffer_rgb










import pyximport; pyximport.install()

from nes.pycore.mos6502 import MOS6502
#from .cycore.mos6502 import MOS6502
from nes.pycore.memory import NESMappedRAM
#from .cycore.memory import NESMappedRAM
from nes.pycore.ppu import NESPPU
#from .cycore.ppu import NESPPU
from nes.rom import ROM
from nes.peripherals import Screen, KeyboardController, ControllerBase
#from nes import LOG_CPU, LOG_PPU, LOG_MEMORY
import pickle

import logging

import pygame

class InterruptListener:
    def __init__(self):
        self._nmi = False
        self._irq = False   # the actual IRQ line on the 6502 rests high and is low-triggered
        self.oam_dma_pause = False

    def raise_nmi(self):
        self._nmi = True

    def reset_nmi(self):
        self._nmi = False

    def reset_oam_dma_pause(self):
        self.oam_dma_pause = False

    def raise_oam_dma_pause(self):
        self.oam_dma_pause = True

    def any_active(self):
        return self._nmi or self._irq or self.oam_dma_pause

    @property
    def nmi_active(self):
        return self._nmi

    @property
    def irq_active(self):
        return self._irq


class NES:
    """
    The NES system itself, combining all of the parts, and the interlinks
    """
    PPU_CYCLES_PER_CPU_CYCLE = 3
    FRAMERATE_FPS = 240

    def __init__(self, rom_file, screen_scale=3, log_file=None, log_level=None, prg_start=None):
        """
        Build a NES and cartridge from the bits and pieces we have lying around plus some rom data!  Also do some things
        like set up logging, etc.
        """
        # set up the logger
        self.init_logging(log_file, log_level)

        rom = ROM(rom_file, py_compatibility_mode=True)

        # the cartridge is a piece of hardware (unlike the ROM, which is just data) and must come first because it
        # supplies bits of hardware (memory in most cases, both ROM and potentially RAM) all over the system, so it
        # affects the actual hardware configuration of the system.
        self.cart = rom.get_cart(prg_start)

        # game controllers have no dependencies
        self.controller1 = KeyboardController()
        self.controller2 = ControllerBase(active=False)   # connect a second gamepad, but make it inactive for now

        # the interrupt listener here is not a NES hardware device, it is an interrupt handler that is used to pass
        # interrupts between the PPU and CPU in this emulator
        self.interrupt_listener = InterruptListener()
        self.ppu = NESPPU(cart=self.cart, interrupt_listener=self.interrupt_listener)

        # screen needs to have the PPU
        self.screen = Screen(ppu=self.ppu, scale=screen_scale, py_compatibility_mode=True)
        self.ppu.screen = self.screen
        self.screen_scale = screen_scale

        # due to memory mapping, lots of things are connected to the main memory
        self.memory = NESMappedRAM(ppu=self.ppu,
                                   apu=None,
                                   cart=self.cart,
                                   controller1=self.controller1,
                                   controller2=self.controller2,
                                   interrupt_listener=self.interrupt_listener
                                   )

        # only the memory is connected to the cpu, all access to other devices is done through memory mapping
        self.cpu = MOS6502(memory=self.memory,
                           undocumented_support_level=2,  # a few NES games make use of some more common undocumented instructions
                           stack_underflow_causes_exception=False
                           )

        # Let's get started!  Reset the cpu so we are ready to go...
        self.cpu.reset()

    def init_logging(self, log_file, log_level):
        """
        Initialize the logging; set the log file and the logging level (LOG_MEMORY, LOG_PPU, LOG_CPU are all below
        logging.DEBUG)
        """
        if log_file is None or log_level is None:
            logging.disable()
            return

        logging.addLevelName(LOG_MEMORY, "MEMORY")  # set a low level for memory logging because it is so intense
        logging.addLevelName(LOG_PPU, "PPU")
        logging.addLevelName(LOG_CPU, "CPU")
        # set up logging to the format as we required
        logging.basicConfig(filename=log_file,
                            level=logging.NOTSET,
                            format='%(asctime)-15s %(source)-5s %(message)s',
                            filemode='w',
                            )
        logging.root.setLevel(log_level)

    def __getstate__(self):
        state = self.memory.__getstate__() #(self.cart, self.controller1, self.controller2, self.interrupt_listener, self.cpu, self.memory, self.screen_scale)
        return state

    def __setstate__(self, state):
        self.cart, self.controller1, self.controller2, self.interrupt_listener, self.cpu, self.memory, self.screen_scale = state
        self.screen = Screen(scale=self.screen_scale)

    def save(self):
        with open("test.p", "wb") as f:
            pickle.dump(self, f, protocol=4)

    @staticmethod
    def load(file):
        with open(file, "rb") as f:
            nes = pickle.load(f)
        nes.screen._special_init()
        return nes

    def step(self):
        """
        The heartbeat of the system.  Run one instruction on the CPU and the corresponding amount of cycles on the
        PPU (three per CPU cycle, at least on NTSC systems).
        """
        if self.interrupt_listener.any_active():
            # do it this way for speed
            if self.interrupt_listener.nmi_active:
                print("NMI Triggered")
                cpu_cycles = self.cpu.trigger_nmi()  # raises an NMI on the CPU, but this does take some CPU cycles
                #print("CPU PC: {:X}".format(self.cpu.PC))

                # should we do this here or leave this up to the triggerer?  It's really up to the triggerer to put the NMI
                # line into its "off" position, but if the line remains in the same state the CPU will not trigger another
                # NMI anyway.  Electronics lend themselves to edge-triggered events, whereas software is more naturally
                # pulse triggered in some ways (a thing happens then returns).
                self.interrupt_listener.reset_nmi()
            elif self.interrupt_listener.irq_active:
                raise NotImplementedError("IRQ is not implemented")
            elif self.interrupt_listener.oam_dma_pause:
                # https://wiki.nesdev.com/w/index.php/PPU_OAM#DMA
                cpu_cycles = self.cpu.oam_dma_pause()
                #cpu_cycles = self.OAM_DMA_CPU_CYCLES + self.cpu.cycles_since_reset % 2
                #self.cpu.cycles_since_reset += cpu_cycles  #todo: should we do this - don't think it matters
                self.interrupt_listener.reset_oam_dma_pause()
        else:
            cpu_cycles = self.cpu.run_next_instr()

        #print(cpu_cycles, self.ppu.line, self.ppu.pixel)
        vblank_started = self.ppu.run_cycles(cpu_cycles * self.PPU_CYCLES_PER_CPU_CYCLE)
        #print(cpu_cycles, self.ppu.line, self.ppu.pixel)
        return vblank_started

    def run(self):
        """
        Run the NES indefinitely (or until some quit signal); this will only exit on quit.
        There is some PyGame specific stuff in here in order to handle frame timing and checking for exits
        """
        pygame.init()
        clock = pygame.time.Clock()

        show_hud = True

        while True:
            vblank_started=False
            while not vblank_started:
                if self.interrupt_listener.any_active():
                    # do it this way for speed
                    if self.interrupt_listener.nmi_active:
                        cpu_cycles = self.cpu.trigger_nmi()  # raises an NMI on the CPU, but this does take some CPU cycles
                        # print("CPU PC: {:X}".format(self.cpu.PC))

                        # should we do this here or leave this up to the triggerer?  It's really up to the triggerer to put the NMI
                        # line into its "off" position, but if the line remains in the same state the CPU will not trigger another
                        # NMI anyway.  Electronics lend themselves to edge-triggered events, whereas software is more naturally
                        # pulse triggered in some ways (a thing happens then returns).
                        self.interrupt_listener.reset_nmi()
                    elif self.interrupt_listener.irq_active:
                        raise NotImplementedError("IRQ is not implemented")
                    elif self.interrupt_listener.oam_dma_pause:
                        # https://wiki.nesdev.com/w/index.php/PPU_OAM#DMA
                        cpu_cycles = self.cpu.oam_dma_pause()
                        # cpu_cycles = self.OAM_DMA_CPU_CYCLES + self.cpu.cycles_since_reset % 2
                        # self.cpu.cycles_since_reset += cpu_cycles  #todo: should we do this - don't think it matters
                        self.interrupt_listener.reset_oam_dma_pause()
                else:
                    cpu_cycles = self.cpu.run_next_instr()

                #print(self.cpu.log_line())

                vblank_started = self.ppu.run_cycles(cpu_cycles * self.PPU_CYCLES_PER_CPU_CYCLE)

            # update the controllers once per frame
            self.controller1.update()
            self.controller2.update()

            fps = clock.get_fps()
            if show_hud:
                self.screen.add_text("{:.0f} fps".format(clock.get_fps()), (10, 10), (0, 255, 0) if fps > 55 else (255, 0, 0))

            # Check for an exit
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    return
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_1:
                        show_hud = not show_hud
                    if event.key == pygame.K_0:
                        self.save()
                        self.screen.add_text("saved", (100, 10), (255, 128, 0))

            self.screen.show()
            clock.tick(self.FRAMERATE_FPS)
            #print("frame end:  {:.1f} fps".format(clock.get_fps()))






"""
TODO:

- IRQ is not implemented (and is used in a few places on the NES:  https://wiki.nesdev.com/w/index.php/IRQ)
- colliding and lost interrupts:  http://visual6502.org/wiki/index.php?title=6502_Timing_of_Interrupt_Handling

"""





# import pyximport; pyximport.install()

# important to cimport these rather than import them because that makes them have a lot less python interaction
from .bitwise cimport bit_high, bit_low, set_bit, clear_bit, set_high_byte, set_low_byte
from .memory cimport NESVRAM, PALETTE_START, NAMETABLE_START, NAMETABLE_LENGTH_BYTES, ATTRIBUTE_TABLE_OFFSET

# A NES rgb palette mapping from NES color values to RGB; others are possible.
cdef int[64][3] DEFAULT_NES_PALETTE = [
    ( 82,  82,  82), (  1,  26,  81), ( 15,  15, 101), ( 35,   6,  99),
    ( 54,   3,  75), ( 64,   4,  38), ( 63,   9,   4), ( 50,  19,   0),
    ( 31,  32,   0), ( 11,  42,   0), (  0,  47,   0), (  0,  46,  10),
    (  0,  38,  45), (  0,   0,   0), (  0,   0,   0), (  0,   0,   0),
    (160, 160, 160), ( 30,  74, 157), ( 56,  55, 188), ( 88,  40, 184),
    (117,  33, 148), (132,  35,  92), (130,  46,  36), (111,  63,   0),
    ( 81,  82,   0), ( 49,  99,   0), ( 26, 107,   5), ( 14, 105,  46),
    ( 16,  92, 104), (  0,   0,   0), (  0,   0,   0), (  0,   0,   0),
    (254, 255, 255), (105, 158, 252), (137, 135, 255), (174, 118, 255),
    (206, 109, 241), (224, 112, 178), (222, 124, 112), (200, 145,  62),
    (166, 167,  37), (129, 186,  40), ( 99, 196,  70), ( 84, 193, 125),
    ( 86, 179, 192), ( 60,  60,  60), (  0,   0,   0), (  0,   0,   0),
    (254, 255, 255), (190, 214, 253), (204, 204, 255), (221, 196, 255),
    (234, 192, 249), (242, 193, 223), (241, 199, 194), (232, 208, 170),
    (217, 218, 157), (201, 226, 158), (188, 230, 174), (180, 229, 199),
    (181, 223, 228), (169, 169, 169), (  0,   0,   0), (  0,   0,   0),
]

cdef class NESPPU:
    """
    NES Picture Processing Unit (PPU), the 2C02

    References:
        [1] Overall reference:  https://wiki.nesdev.com/w/index.php/PPU_programmer_reference
        [2] Rendering timing: https://wiki.nesdev.com/w/index.php/PPU_rendering
        [3] OAM layout:  https://wiki.nesdev.com/w/index.php/PPU_OAM

        [4] Detailed operation: http://nesdev.com/2C02%20technical%20operation.TXT

        [5] Palette generator: https://bisqwit.iki.fi/utils/nespalette.php

        [6] Register behaviour: https://wiki.nesdev.com/w/index.php/PPU_registers
        [7] Scrolling: https://wiki.nesdev.com/w/index.php/PPU_scrolling
    """
    def __init__(self, cart=None, interrupt_listener=None, palette=None):

        # Registers
        self.ppu_ctrl = 0
        self.ppu_mask = 0
        self.oam_addr = 0
        self.oam_data = 0
        self.ppu_scroll[:] = [0, 0]     # this contains x-scroll (byte 0) and y-scroll (byte 1) accumulated over two writes
        self.ppu_addr = 0               # the accumulated **16-bit** address
        self._ppu_byte_latch = 0        # latch to keep track of which byte is being written in ppu_scroll and ppu_addr; latch is shared

        # internal latches to deal with open bus and buffering behaviour
        self._ppu_data_buffer = 0       # data to hold buffered reads from VRAM (see read of ppu_data)
        self._io_latch = 0              # last write/valid read of the ppu registers, sometimes reflected in read status

        # internal latches used in background rendering
        self._palette = [[0, 1, 2, 3], [4, 5, 6, 7]]
        self._pattern_lo = 0   # 16 bit patterns register to hold 2 x 8 bit patterns
        self._pattern_hi = 0   # 16 bit patterns register to hold 2 x 8 bit patterns

        # internal memory and latches used in sprite rendering
        self._sprite_pattern = [[0] * 8 for _ in range(8)]
        self._sprite_bkg_priority = [0] * 8
        self._num_active_sprites = 0

        # some state used in rendering to tell us where on the screen we are drawing
        self.line = 0
        self.pixel = 0

        # internal statuses
        self.in_vblank = False
        self.sprite_zero_hit = False
        self.sprite_overflow = False
        self.ignore_ppu_ctrl = True   # on startup, ignore writes to ppu_ctrl for about 30k cycles

        # status used by emulator
        self.cycles_since_reset = 0
        self.frames_since_reset = 0  # need all three counters (not really, but easier) because frame lengths vary

        # memory
        self.vram = NESVRAM(cart=cart)

        # interrupt listener
        self.interrupt_listener = interrupt_listener

        # palette: use the default, but can be replaced using utils.load_palette
        self.set_hex_palette(palette if palette is not None else DEFAULT_NES_PALETTE)
        self.transparent_color = -1

        self._palette_cache_valid[:] = [0, 0, 0, 0, 0, 0, 0, 0]
        self.irq_tick_triggers[:] = [False] * 68

    def set_hex_palette(self, rgb_palette):
        for i, c in enumerate(rgb_palette):
            self.hex_palette[i] = (c[0] << 16) + (c[1] << 8) + c[2]

    ################ Copy OAM data #####################################################################################

    cdef void write_oam(self, unsigned char* data):
        cdef int i=0
        for i in range(OAM_SIZE_BYTES):
            self.oam[i] = data[i]

    ################ Register reading and writing ######################################################################

    cdef unsigned char ppu_status(self):
        """
        The ppu status register value (without io latch noise in lower bits)
        :return:
        """
        return VBLANK_MASK * self.in_vblank \
               + SPRITE0_HIT_MASK * self.sprite_zero_hit \
               + SPRITE_OVERFLOW_MASK * self.sprite_overflow

    cdef unsigned char read_register(self, int register):
        """
        Read the specified PPU register (and take the correct actions along with that)
        This is mostly (always?) triggered by the CPU reading memory mapped ram at 0x2000-0x3FFF
        "Reading a nominally write-only register will return the latch's current value" [6].
        The "latch" here refers to the capacitance
        of the PPU lines, which leads to some degree of "memory" on the lines, which will hold the last
        value written to a port (including read only ones), or the last value read from a read-only port
        """

        cdef unsigned char v=0

        if register == PPU_CTRL:
            # write only
            return self._io_latch
        elif register == PPU_MASK:
            # write only
            return self._io_latch
        elif register == PPU_STATUS:
            # clear ppu_scroll and ppu_addr latches
            self._ppu_byte_latch = 0   # this is a shared latch between scroll and addr
            v = self.ppu_status() + (0x00011111 & self._io_latch)
            self.in_vblank = False  # clear vblank in ppu_status
            self._io_latch = v
            return v
        elif register == OAM_ADDR:
            # write only
            return self._io_latch
        elif register == OAM_DATA:
            # todo: does not properly implement the weird results of this read during rendering
            v = self.oam[self.oam_addr]
            self._io_latch = v
            return v
        elif register == PPU_SCROLL:
            # write only
            return self._io_latch
        elif register == PPU_ADDR:
            # write only
            return self._io_latch
        elif register == PPU_DATA:
            if self.ppu_addr < PALETTE_START:
                v = self._ppu_data_buffer
                self._ppu_data_buffer = self.vram.read(self.ppu_addr)
            else:
                v = self.vram.read(self.ppu_addr)
                # palette reads will return the palette without buffering, but will put the mirrored NT byte in the
                # read buffer, i.e. reading $3F00 will give you the palette entry at $3F00 and will put the byte in
                # VRAM[$2F00] in the read buffer
                # source: http://forums.nesdev.com/viewtopic.php?t=1721
                self._ppu_data_buffer = self.vram.read(self.ppu_addr - 0x1000)
            self._increment_vram_address()
            self._io_latch = self._ppu_data_buffer
            return v

    cdef void write_register(self, int register, unsigned char value):
        """
        Write one of the PPU registers with byte value and do whatever else that entails
        """
        # need to store the last write because it affects the value read on ppu_status
        # "Writing any value to any PPU port, even to the nominally read-only PPUSTATUS, will fill this latch"  [6]
        cdef int trigger_nmi=0, n_chng=0, prev_ppu_addr=0
        value &= 0xFF  # can only write a byte here
        self._io_latch = value

        if register == PPU_CTRL:
            # write only
            # writes to ppu_ctrl are ignored at first
            # this is a slightly safer mechanism to implement this rather than relying on cycles_since_reset, which
            # caused a bug by overflowing
            if self.ignore_ppu_ctrl:
                if self.cycles_since_reset < 29658:
                    return
                else:
                    self.ignore_ppu_ctrl = False
            # can trigger an immediate NMI if we are in vblank and the (allow) vblank NMI trigger flag is flipped high
            trigger_nmi = self.in_vblank \
                          and (value & VBLANK_MASK) > 0 \
                          and (self.ppu_ctrl & VBLANK_MASK) == 0
            self.ppu_ctrl = value
            if trigger_nmi:
                self._trigger_nmi()
        elif register == PPU_MASK:
            # write only
            self.ppu_mask = value
        elif register == PPU_STATUS:
            # read only
            pass
        elif register == OAM_ADDR:
            # write only
            self.oam_addr = value
        elif register == OAM_DATA:
            # read/write
            self.oam[self.oam_addr] = value
            # increment the OAM address and wrap around if necessary (wraps to start of page since OAM addr specifies
            # sub-page address)
            self.oam_addr = (self.oam_addr + 1) & 0xFF
        elif register == PPU_SCROLL:
            # write only
            self.ppu_scroll[self._ppu_byte_latch] = value
            # flip which byte is pointed to on each write; reset on ppu status read.  Latch shared with ppu_addr.
            self._ppu_byte_latch = 1 - self._ppu_byte_latch
            #print("set scroll: line {}, px {}, y-scroll: {}".format(self.line, self.pixel, self.ppu_scroll[PPU_SCROLL_Y]))
        elif register == PPU_ADDR:
            # write only
            # high byte first
            if self._ppu_byte_latch == 0:
                prev_ppu_addr = self.ppu_addr
                self.ppu_addr = (self.ppu_addr & 0x00FF) + (value << 8)
                # Writes here overwrite the current nametable bits in ppu_ctrl (or at least overwrite bits in an
                # internal latch that is equivalent to this); see [7].  Some games, e.g. SMB, rely on this behaviour
                self.ppu_ctrl = (self.ppu_ctrl & 0b11111100) + ((value & 0b00001100) >> 2)

                # a write here also has a very unusual effect on the coarse and fine y scroll [7]
                # (note that bit 2 of PPU_SCROLL_Y is always set to 0 by this)
                self.ppu_scroll[PPU_SCROLL_Y] = (self.ppu_scroll[PPU_SCROLL_Y] & 0b00111000) + \
                                                     ((value & 0b00000011) << 6) + ((value & 0b00110000) >> 4)
                if bit_low(prev_ppu_addr, 12) and bit_high(self.ppu_addr, 12):
                    self.vram.cart.irq_tick()

                #print("set scroll: line {}, px {}, y-scroll: {}".format(self.line, self.pixel, self.ppu_scroll[PPU_SCROLL_Y]))
            else:
                self.ppu_addr = (self.ppu_addr & 0xFF00) + value
                # writes here have a weird effect on the x and y scroll values [7]
                self.ppu_scroll[PPU_SCROLL_X] = (self.ppu_scroll[PPU_SCROLL_X] & 0b00000111) + ((value & 0b00011111) << 3)
                self.ppu_scroll[PPU_SCROLL_Y] = (self.ppu_scroll[PPU_SCROLL_Y] & 0b11000111) + ((value & 0b11100000) >> 2)

                # side effect is that y-scroll is changed immediately, allowing mid-frame y-scroll changes
                # this only happens on the second write to ppu_addr
                self._reset_effective_y()
                #print("set scroll: line {}, px {}, y-scroll: {} (IMMEDIATE)".format(self.line, self.pixel, self.ppu_scroll[PPU_SCROLL_Y]))

            # flip which byte is pointed to on each write; reset on ppu status read
            self._ppu_byte_latch = 1 - self._ppu_byte_latch
        elif register == PPU_DATA:
            # read/write
            #print("write {:02X} to {:04X}".format(value, self.ppu_addr))
            self.vram.write(self.ppu_addr, value)
            self._increment_vram_address()
            # invalidate the palette cache if this is a write to the palette memory.  Could be more careful about what
            # is invalidated here so that potentially less recalculation is needed.
            if self.ppu_addr >= PALETTE_START:
                self.invalidate_palette_cache()


    ################ Screen buffer copy and clear ######################################################################

    cdef int get_background_color(self):
        """
        Get the background color for the screen (set by color zero of palette zero)
        :return:
        """
        cdef int p0[4]
        self.decode_palette(p0, 0, False)
        return self.hex_palette[p0[0]]

    cdef void _clear_to_bkg(self):
        """
        Clears the screen buffer to the background color (which is set by palette 0)
        :return:
        """
        cdef int cc=0, x=0, y=0

        cc = self.get_background_color()
        for x in range(SCREEN_WIDTH_PX):
            for y in range(SCREEN_HEIGHT_PX):
                self.screen_buffer[x][y] = cc

    cpdef void copy_screen_buffer_to(self, unsigned int[:, :] dest, bint v_overscan=False, bint h_overscan=False):
        """
        Copy the screen buffer to a supplied destination.  Up to the caller to ensure that the destination buffer
        has sufficient space to write into.
        """
        cdef int start_line=VERTICAL_OVERSCAN_PX, end_line=SCREEN_HEIGHT_PX - VERTICAL_OVERSCAN_PX
        cdef int start_row=HORIZONTAL_OVERSCAN_PX, end_row=SCREEN_WIDTH_PX - HORIZONTAL_OVERSCAN_PX
        if v_overscan:
            start_line = 0
            end_line = SCREEN_HEIGHT_PX
        if h_overscan:
            start_row = 0
            end_row = SCREEN_WIDTH_PX
        # create a memory view to the screen to allow it to be treated as a buffer in the Numpy-esque style
        cdef unsigned int[:, :] scr_mv = <unsigned int[:SCREEN_WIDTH_PX, :SCREEN_HEIGHT_PX]>self.screen_buffer
        dest[:, :] = scr_mv[start_row:end_row, start_line:end_line]


    ################ Main cycle running ################################################################################

    cdef int run_cycles(self, int num_cycles):
        """
        The main function of the PPU that steps it forward the specified number of cycles.  Cycles correspond to screen
        pixels during the screen-drawing phase of the ppu there are three ppu cycles per cpu cycles, at least on NTSC
        systems.
        :param num_cycles: the number of PPU cycles (not cpu cycles!).
        :return: whether vblank started during the cycles run
        """
        cdef int cyc=0
        cdef bint vblank_started=False

        for cyc in range(num_cycles):
            # current scanline of the frame we are on - this determines behaviour during the line
            if self.line <= 239 and (self.ppu_mask & RENDERING_ENABLED_MASK) > 0:
                self.render_visible_scanline()
            elif self.line == 241 and self.pixel == 1:
                # set vblank flag
                vblank_started = True   # this is used by the emulator to know when it can flip the screen
                self.in_vblank = True   # set the vblank flag in ppu_status register
                # trigger NMI (if NMI is enabled)
                if (self.ppu_ctrl & VBLANK_MASK) > 0:
                    self._trigger_nmi()
            elif self.line == 261:
                # line 261 is also sometimes called line -1 and is the pre-render scanline
                self.prerender_scanline()

            self.increment_pixel()

        return vblank_started

    cdef void prerender_scanline(self):
        """
        The pre-render scanline (line 261) has some special things that happen on it like turning off vblank and
        starting to do the background pre-fetch.
        """
        if self.pixel == 1:
            # At dot 1, reset vblank flag in ppu_status
            self.in_vblank = False
            self.sprite_zero_hit = False
            self.sprite_overflow = False
        elif self.pixel == 257:
            # "Sprite evaluation does not happen on the pre-render scanline. Because evaluation applies to the
            # next line's sprite rendering, no sprites will be rendered on the first scanline, and this is why
            # there is a 1 line offset on a sprite's Y coordinate."
            # source: https://wiki.nesdev.com/w/index.php/PPU_sprite_evaluation  (note 1)
            self._num_active_sprites = 0
            # reset the x counter to the start of the line
            self._effective_x = self.ppu_scroll[PPU_SCROLL_X] + (bit_high(self.ppu_ctrl, BIT_NAMETABLE_X) << 8)
        elif self.pixel == 260 and (self.ppu_mask & RENDERING_ENABLED_MASK) > 0:
            # for simplicity, always trigger the irq_tick once at pixel 260 on the pre-render scanline; this might be
            # wrong by a few pixels for this line, but it probably won't matter most of the time
            # todo: more correct behaviour here (what is it in the case of double-height sprites?)
            self.vram.cart.irq_tick()
        elif self.pixel == 280:
            # "Vertical scroll bits are reloaded if rendering is enabled"
            self._reset_effective_y()
        elif self.pixel == 328 or self.pixel == 336:
            # load background data for next scanline
            self.fill_bkg_latches()  # get some more data for the upper latches
        elif self.pixel == 339:
            # about the last pixel of the pre-render, so clear the display ready for the next frame
            self._clear_to_bkg()

    cdef void render_visible_scanline(self):
        """
        Render a pixel on a visible scanline (lines 0-239 inclusive)
        """
        cdef int bkg_pixel=0, final_pixel=0, coarse_y=0, cmask=0, sprite_table=0
        cdef bint double_sprites=False

        if 0 < self.pixel <= 256:  # pixels 1 - 256
            # render pixel - 1
            if (self.pixel - 1) % 8 == 0 and self.pixel > 1:
                # fill background data latches
                # todo: this is not cycle-correct, since the read is done atomically at the eighth pixel rather than throughout the cycle.
                self.fill_bkg_latches()   # get some more data for the upper latches
            # render background from latches
            bkg_pixel = self._get_bkg_pixel()
            # overlay sprite from latches
            final_pixel = self._overlay_sprites(bkg_pixel)
            # monochrome mode if bit 0 of ppu_mask is set
            cmask = 0x30 if (self.ppu_mask & 1) else 0xFF
            if final_pixel != self.transparent_color:
                self.screen_buffer[self.pixel - 1][self.line] = self.hex_palette[final_pixel & cmask]
        elif self.pixel == 257:   # pixels 257 - 320
            # sprite data fetching: fetch data from OAM for sprites on the next scanline
            # NOTE:  "evaluation applies to the next line's sprite rendering, ... and this is why
            # there is a 1 line offset on a sprite's Y coordinate."
            # source: https://wiki.nesdev.com/w/index.php/PPU_sprite_evaluation  (note 1)
            self._prefetch_active_sprites()
            # reset the x counter to the start of the line
            self._effective_x = self.ppu_scroll[PPU_SCROLL_X] + (bit_high(self.ppu_ctrl, BIT_NAMETABLE_X) << 8)
        elif self.pixel == 280:
            # increment y to next row
            if (self._effective_y & 0xFF) == 239:  # can only go up to 239 before wrapping to 256
                self._effective_y += 17  # skip over the attribute table to the next nametable (equivalent to 2 rows)
            elif (self._effective_y & 0xFF) == 255:  # but if scroll is set into the "illegal zone"
                # if scroll is set into the out-of-bounds zone between lines 240-255, that is okay, which causes the
                # attribute table to be read as tile data; this achieves a sort of negative scroll that some games use
                # especially noticable in the bouncing of the SMB3 logo in the title sequence of SMB3, which will not
                # work properly without this obscure behaviour. See https://wiki.nesdev.com/w/index.php/PPU_scrolling
                self._effective_y &= 0x100
            else:
                self._effective_y += 1
        elif self.pixel == 328 or self.pixel == 336:   # pixels 321 - 336
            # fill background data latches with data for first two tiles of next scanline
            self.fill_bkg_latches()  # get some more data for the upper latches
        else:  # pixels 337 - 340
            # todo: garbage nametable fetches (used by MMC5)
            pass

        if 257 <= self.pixel <= 324:  # sprite fetching period
            # sprite fetch has already happened on pixel 257 (above), but it is possible that, if using double sprites
            # irq ticks could be triggered here; the timings of these were precalculated during sprite fetch and put
            # into the irq_tick_triggers array, so all we have to do here is test if this pixel has the trigger be true
            if self.irq_tick_triggers[self.pixel - 257]:
                self.vram.cart.irq_tick()

    cdef void increment_pixel(self):
        """
        Increment the current pixel by 1 and, if necessary, move to the next line or even frame.
        """
        self.cycles_since_reset += 1
        self.pixel += 1
        # need to check if we are in the last *two* pixels because the frame sometimes ends one pixel early
        if self.pixel >= PIXELS_PER_LINE - 1:
            if self.line == 261 and self.pixel == PIXELS_PER_LINE - self.frames_since_reset % 2:
                # frame has ended
                self.pixel = 0
                self.line = 0
                self.frames_since_reset += 1
            elif self.pixel == PIXELS_PER_LINE:
                # just this line has ended
                self.line += 1
                self.pixel = 0

    cdef void _increment_vram_address(self):
        """
        Increment vram address after reads/writes by an amount specified by a value in ppu_ctrl
        """
        cdef int prev_ppu_addr = self.ppu_addr
        self.ppu_addr += 1 if (self.ppu_ctrl & VRAM_INCREMENT_MASK) == 0 else 32
        if bit_low(prev_ppu_addr, 12) and bit_high(self.ppu_addr, 12):
            # if line A12 changes from low to high, trigger an irq counter tick in the cartridge (MMC3 and derivatives)
            self.vram.cart.irq_tick()

    cdef void _trigger_nmi(self):
        """
        Do whatever is necessary to trigger an NMI to the CPU; note that it is up to the caller to check whether NMIs
        should be generated by the PPU at this time (a flag in ppu_ctrl), and respecting this is critically important.
        """
        self.interrupt_listener.raise_nmi()

    cdef void _reset_effective_y(self):
        # slightly complicated by the 30 (x 8px) row height of the screen, so if y >= 240, should bump it by 16 to
        # account for the fact that the attribute table has been crossed.
        self._effective_y = (self.ppu_scroll[PPU_SCROLL_Y] + (bit_high(self.ppu_ctrl, BIT_NAMETABLE_Y) << 8)) & 0x1FF
        # it is actually valid to set the y scroll to between 240 and 256, will cause attriute table to be read as
        # nametable data; used in some places to do "negative y-scroll"
        #if self.ppu_scroll[PPU_SCROLL_Y] >= 240:
        #   self._effective_y += 16


    ################ Sprite decoding and rendering #####################################################################

    cdef void _prefetch_active_sprites(self):
        """
        Non cycle-correct detector for active sprites on the given line.  Returns a list of the indices of the start
        address of the sprite in the OAM
        """
        cdef int n=0, addr=0, sprite_y=0
        cdef list sprite_line

        # scan through the sprites, starting at oam_start_addr, seeing if they are visible in the line given
        # (note that should be the next line); if so, add them to the list of active sprites, until that gets full.
        # if using 8x16 sprites (True), or 8x8 sprite (False)
        cdef bint double_sprites = (self.ppu_ctrl & SPRITE_SIZE_MASK) > 0
        cdef int sprite_height = 16 if double_sprites else 8

        self._num_active_sprites = 0
        for n in range(64):
            addr = (self.oam_addr + n * 4) & 0xFF  # wrap around the address if need be
            sprite_y = self.oam[addr]
            if sprite_y <= self.line < sprite_y + sprite_height:
                if self._num_active_sprites < 8:
                    self._active_sprite_addrs[self._num_active_sprites] = addr
                    self._sprite_line[self._num_active_sprites] = self.line - sprite_y
                    self._num_active_sprites += 1
                    #sprite_line.append(line - sprite_y)
                else:
                    # overflow
                    # todo: this implements the *correct* behaviour of sprite overflow, but not the buggy behaviour
                    # (and even then it is not cycle correct, so could screw up games that rely on timing of this very exactly)
                    self.sprite_overflow = True
                    break
        self._fill_sprite_latches(double_sprites)

    cdef void _fill_sprite_latches(self, int double_sprites):
        """
        Non cycle-correct way to pre-fetch the sprite lines for the next scanline
        """
        cdef int table_base=0, prev_table_base=0, palette_ix=0, attribs=0, flip_v=0, flip_h=0, tile_ix=0, line=0, tile_base=0, x=0, c=0
        cdef int palette[4]
        cdef unsigned char sprite_pattern_lo=0, sprite_pattern_hi=0
        cdef int i=0, address=0

        # previously the ppu was fetching background tiles, so address line A12 was set by the bkg table base; this is
        # used for triggering irq ticks for MMC3 type cartridges.
        table_base = ((self.ppu_ctrl & BKG_PATTERN_TABLE_MASK) > 0) * 0x1000

        # the IRQ tick triggers for MMC3 are handled on the scanline renderers themselves, but they are related to
        # the rendering of sprites and associated pattern table switching, so they are calculated here
        self.irq_tick_triggers[:] = [False] * 68
        if not double_sprites:
            if table_base == 0:
                self.irq_tick_triggers[67] = True   # trigger once at pixel 324
            else:
                self.irq_tick_triggers[3] = True    # trigger once at pixel 260

        for i in range(self._num_active_sprites):
            address = self._active_sprite_addrs[i]

            attribs = self.oam[(address + 2) & 0xFF]
            palette_ix = attribs & 0b00000011

            self.decode_palette(palette, palette_ix, is_sprite=True)
            flip_v = bit_high(attribs, bit=7)
            flip_h = bit_high(attribs, bit=6)
            self._sprite_bkg_priority[i] = bit_high(attribs, bit=5)

            if not double_sprites:
                # sprite table select bit is only used in single-height (8x8) mode
                table_base = ((self.ppu_ctrl & SPRITE_PATTERN_TABLE_MASK) > 0) * 0x1000
                tile_ix = self.oam[(address + 1) & 0xFF]
                line = self._sprite_line[i] if not flip_v else 7 - self._sprite_line[i]
            else:
                line = self._sprite_line[i] if not flip_v else 15 - self._sprite_line[i]
                tile_ix = self.oam[(address + 1) & 0xFF] & 0b11111110
                # in double-height mode, sprite table is set by bottom bit of 2nd byte in OAM
                prev_table_base = table_base
                table_base = (self.oam[(address + 1) & 0xFF] & 1) * 0x1000
                if prev_table_base==0 and table_base > 0:
                    # rising edge on the table selecting address line triggers the irq_counter
                    self.irq_tick_triggers[i * 8] = True

                if line >= 8:
                    # in the lower tile
                    tile_ix += 1
                    line -= 8

            tile_base = table_base + tile_ix * PATTERN_SIZE_BYTES
            sprite_pattern_lo = self.vram.read(tile_base + line)
            sprite_pattern_hi = self.vram.read(tile_base + 8 + line)

            for x in range(8):
                c = bit_high(sprite_pattern_hi, x) * 2 + bit_high(sprite_pattern_lo, x)
                self._sprite_pattern[i][x if flip_h else 7 - x] = palette[c] if c else self.transparent_color

        if double_sprites and self._num_active_sprites < 8:
            # at this point, the PPU does some reads from table 1 that are discarded, but that means that if the table
            # base was zero, it will then go high, triggering an irq tick
            if table_base == 0:
                self.irq_tick_triggers[self._num_active_sprites * 8] = True

    cdef int _overlay_sprites(self, int bkg_pixel):
        """
        Cycle-correct (ish) sprite rendering for the pixel at y=line, pixel=pixel.  Includes sprite 0 collision detection.
        """
        cdef int sprite_c_out=0, c=0, top_sprite=0, sprite_addr=0, sprite_x=0, pix=0
        cdef bint s0_visible=False

        c_out = bkg_pixel
        if (self.ppu_mask & RENDER_SPRITES_MASK) == 0 \
            or (self.pixel - 1 < 8 and bit_low(self.ppu_mask, RENDER_LEFT8_SPRITES_BIT)):
            return c_out

        sprite_c_out = self.transparent_color
        top_sprite = -1
        for i in range(self._num_active_sprites - 1, -1, -1):  # iterate from backmost to frontmost sprite #reversed(range(self._num_active_sprites)):
            # render in reverse to make overwriting easier
            sprite_addr = self._active_sprite_addrs[i]
            sprite_x = self.oam[sprite_addr + 3]
            if sprite_x <= self.pixel - 1 < sprite_x + 8:
                pix = self.pixel - 1 - sprite_x
                # this sprite is visible now
                c = self._sprite_pattern[i][pix]
                if c != self.transparent_color:
                    top_sprite = i
                    sprite_c_out = c
                    if sprite_addr == 0:
                        s0_visible = True

        # sprite zero collision detection
        # Details: https://wiki.nesdev.com/w/index.php/PPU_OAM#Sprite_zero_hits
        if not self.sprite_zero_hit and s0_visible and bkg_pixel != self.transparent_color:
            # todo: there are some more fine details here
            #print("s0 hit ", self.line, self.pixel)
            self.sprite_zero_hit = True

        # now decide whether to keep sprite or bkg pixel
        if sprite_c_out != self.transparent_color and (not self._sprite_bkg_priority[top_sprite] or bkg_pixel == self.transparent_color):
            c_out = sprite_c_out

        return c_out #if c_out != self.transparent_color else self.bkg_color  # background color


    ################ Background rendering ##############################################################################

    cdef void fill_bkg_latches(self):
        """
        Fill the ppu's rendering latches with the next tile to be rendered.  Relies on the correct values in
        the internal _effective_x and _effective_y variables.
        """
        cdef int ntbl_base=0, tile_addr=0, tile_line=0, tile_index=0, tile_bank=0, tile_base=0
        cdef int i=0, shift=0, palette_id=0, table_base=0, attr_addr=0
        cdef unsigned char attribute_byte=0, mask=0

        # which nametable are we currently on?
        ntbl_base = (NAMETABLE_START                                  # nametable start
                     + (((self._effective_y >> 8) & 1) << 11)         # y nametable bit
                     + (((self._effective_x >> 8) & 1) << 10)         # x nametable bit
                     )

        ##### Tile reading

        tile_addr = (ntbl_base
                     + (((self._effective_y >> 3) & 0b11111) << 5)    # coarse y   { & 0b11111 necessary because of
                     + ((self._effective_x >> 3) & 0b11111)           # coarse x   { nametable carry in effective x/y
                     )

        tile_line = self._effective_y & 0b111                         # offset within the tile

        # read the tile id from the nametable
        tile_index = self.vram.read(tile_addr)
        # now figure out where the data for that tile is located in the memory
        tile_bank = (self.ppu_ctrl & BKG_PATTERN_TABLE_MASK) > 0
        table_base = tile_bank * 0x1000
        tile_base = table_base + tile_index * PATTERN_SIZE_BYTES

        # shift up the lower bits of the background latches ready to be refilled
        self._pattern_hi <<= 8
        self._pattern_lo <<= 8
        # then read the new lower pattern byte
        self._pattern_lo = (self._pattern_lo & 0xFF00) + self.vram.read(tile_base + tile_line)
        self._pattern_hi = (self._pattern_hi & 0xFF00) + self.vram.read(tile_base + tile_line + 8)

        ##### Attribute (palette) for this block

        attr_addr = (ntbl_base
                     + ATTRIBUTE_TABLE_OFFSET                         # go to the attribute table
                     + (((self._effective_y >> 5) & 0b111) << 3)      # top 3 bits of coarse y  (== coarse y / 4)
                     + ((self._effective_x >> 5) & 0b111)             # top 3 bits of coarse x  (== coarse x / 4)
                     )

        attribute_byte = self.vram.read(attr_addr)

        shift = (((self._effective_y >> 4) & 1) * 4                   # 4th bit of effective_y gives 2-tile block
                 + ((self._effective_x >> 4) & 1) * 2)                # 4th bit of effective_x gives 2-tile block
        mask = 0b00000011 << shift
        palette_id = (attribute_byte & mask) >> shift

        for i in range(4):
            self._palette[0][i] = self._palette[1][i]
        self.decode_palette(self._palette[1], palette_id, False)

        # after a read, increment the effective x so that we can read the next tile along next time
        self._effective_x += 8

    cdef int _get_bkg_pixel(self):
        """
        Determine the current background pixel to draw based on the current internal state of the PPU
        :return:
        """
        cdef int fine_x=0, px=0, v=0, mask=0

        if (   self.ppu_mask & RENDER_BACKGROUND_MASK) == 0 \
            or (self.pixel - 1 < 8 and bit_low(self.ppu_mask, RENDER_LEFT8_BKG_BIT)):
            # in this case, background rendering is off for this pixel, so just return
            return self.transparent_color

        fine_x = self.ppu_scroll[PPU_SCROLL_X] & 0b00000111
        px = (self.pixel - 1) % 8 + fine_x
        mask = 1 << (15 - px)
        v = ((self._pattern_lo & mask) > 0) + ((self._pattern_hi & mask) > 0) * 2
        return self._palette[px >> 3][v] if v > 0 else self.transparent_color


    ################ Palette Decoder ###################################################################################

    cdef void invalidate_palette_cache(self):
        """
        Invalidates the entire palette cache; could be more efficient by only invalidating entries that have been
        rewritten.
        """
        cdef int i=0
        for i in range(8):
            self._palette_cache_valid[i] = False

    cdef void decode_palette(self, int* palette_out, int palette_id, bint is_sprite):
        """
        If is_sprite is true, then decodes palette from the sprite palettes, otherwise
        decodes from the background palette tables.
        """
        cdef int palette_address, i, cmask

        if self._palette_cache_valid[is_sprite * 4 + palette_id]:
            palette_out[0] = self._palette_cache[is_sprite * 4 + palette_id][0]
            palette_out[1] = self._palette_cache[is_sprite * 4 + palette_id][1]
            palette_out[2] = self._palette_cache[is_sprite * 4 + palette_id][2]
            palette_out[3] = self._palette_cache[is_sprite * 4 + palette_id][3]

        # get the palette colours (these are in hue (chroma) / value (luma) format.)
        # palette_id is in range 0..3, and gives an offset into one of the four background palettes,
        # each of which consists of three colors, each of which is represented by a singe byte
        palette_address = PALETTE_START + 16 * is_sprite + 4 * palette_id
        for i in range(4):
            palette_out[i] = (self.vram.read(palette_address + i) & 0b00111111)
            self._palette_cache[is_sprite * 4 + palette_id][i] = palette_out[i]
        self._palette_cache_valid[is_sprite * 4 + palette_id] = True


    ################ Logging ###########################################################################################

    def log_line(self):
        """
        Generate a log line from the CPU
        """
        log = "{:5d}, {:3d}, {:3d}   ".format(self.frames_since_reset, self.line, self.pixel)
        log += "C:{:02X} M:{:02X} S:{:02X} OA:{:02X} OD:{:02X} ".format(self.ppu_ctrl,
                                                                        self.ppu_mask,
                                                                        self.ppu_status,
                                                                        self.oam_addr,
                                                                        self.oam_data)

        log += "SC:{:02X},{:02X} PA:{:04X}".format(self.ppu_scroll[0],
                                                   self.ppu_scroll[1],
                                                   self.ppu_addr)
        return log

    ################ Debug #############################################################################################

    cpdef void debug_render_nametables(self, unsigned int[:, :] dest):
        """
        Reads the nametable and attribute table and then sends the result of that for each
        tile on the screen to render_tile to actually render the tile (reading the pattern tables, etc.)
        """
        cdef int nx=0, ny=0, nametable=0, addr_base=0, row=0, vblock=0, v_subblock_ix=0, col=0, tile_index=0, hblock=0, h_subblock_ix=0, shift=0
        cdef int mask=0, palette_id=0, tile_bank=0, x=0, y=0, x0=0, xe=0, y0=0, ye=0
        cdef unsigned char attribute_byte=0

        dest[:, :] = self.get_background_color()

        for ny in range(2):
            for nx in range(2):
                nametable = ny * 2 + nx
                addr_base = NAMETABLE_START + nametable * NAMETABLE_LENGTH_BYTES
                for row in range(SCREEN_TILE_ROWS):
                    vblock = int(row / 2)
                    v_subblock_ix = vblock % 2
                    for col in range(SCREEN_TILE_COLS):
                        tile_index = self.vram.read(addr_base + row * SCREEN_TILE_COLS + col)

                        # get attribute byte from the attribute table at the end of the nametable
                        # these tables compress the palette id into two bits for each 2x2 tile block of 16x16px.
                        # the attribute bytes each contain the palette id for a 2x2 block of these 16x16 blocks
                        # so get the correct byte and then extract the actual palette id from that
                        hblock = int(col / 2)
                        attribute_byte = self.vram.read(addr_base
                                                        + ATTRIBUTE_TABLE_OFFSET
                                                        + (int(vblock / 2) * 8 + int(hblock / 2))
                                                       )
                        h_subblock_ix = hblock % 2
                        shift = 4 * v_subblock_ix + 2 * h_subblock_ix
                        mask = 0b00000011 << shift
                        palette_id = (attribute_byte & mask) >> shift
                        # ppu_ctrl tells us whether to read the left or right pattern table, so let's fetch that
                        tile_bank = (self.ppu_ctrl & BKG_PATTERN_TABLE_MASK) > 0
                        tile = self.debug_render_tile(dest,
                                                      (nx * SCREEN_TILE_COLS + col) * TILE_WIDTH_PX,
                                                      (ny * SCREEN_TILE_ROWS + row) * TILE_HEIGHT_PX,
                                                      tile_index,
                                                      tile_bank,
                                                      -1,
                                                      #palette_id,
                                                      False,
                                                      False
                                                     )
        # draw a box around the active window
        x0 = (self.ppu_scroll[PPU_SCROLL_X] + (bit_high(self.ppu_ctrl, BIT_NAMETABLE_X) << 8)) % 512
        y0 = (self.ppu_scroll[PPU_SCROLL_Y] + (bit_high(self.ppu_ctrl, BIT_NAMETABLE_Y) * 240)) % 480

        xe = (x0 + 256) % 512
        ye = (y0 + 240) % 480

        dest[x0:min(512, x0 + 256), y0] = 255 << 8
        dest[x0:min(512, x0 + 256), ye] = 255 << 8
        if x0 + 256 > 512:
            dest[0:xe, y0] = 255 << 8
            dest[0:xe, ye] = 255 << 8

        dest[x0, y0:min(480, y0 + 240)] = 255 << 8
        dest[xe, y0:min(480, y0 + 240)] = 255 << 8
        if y0 + 240 > 480:
            dest[x0, 0:ye] = 255 << 8
            dest[xe, 0:ye] = 255 << 8


    cpdef void debug_render_tile(self, unsigned int[:, :] dest, int x0, int y0, int tile_index, int tile_bank, int palette_id, bint flip_h, bint flip_v):
        """
        Decodes a tile given by tile_index from the pattern table specified by tile_bank to an array of RGB color value,
        using the palette supplied.  Transparent pixels (value 0 in the tile) are replaced with self.transparent_color.
        This makes them ready to be blitted to the screen.
        """
        cdef int x=0, y=0, xx=0, yy=0, pixel_color_ix=0, plane=0, table_base=0, tile_base=0
        cdef int palette[4]

        if palette_id > 0:
            self.decode_palette(palette, palette_id, is_sprite=False)
        else:
            palette[0] = 0x01
            palette[1] = 0x11
            palette[2] = 0x21
            palette[3] = 0x31

        # now decode the tile
        table_base = tile_bank * 0x1000

        # tile index tells us which pattern table to read
        tile_base = table_base + tile_index * PATTERN_SIZE_BYTES

        # the (palettized) tile is stored as 2x8byte bit planes, each representing an 8x8 bitmap of depth 1 bit
        # tile here is indexed tile[row][column], *not* tile[x][y]
        #tile = [[0] * TILE_WIDTH_PX for _ in range(TILE_HEIGHT_PX)]

        for y in range(TILE_HEIGHT_PX):
            for x in range(TILE_WIDTH_PX):
                xx = x if not flip_h else TILE_WIDTH_PX - 1 - x
                yy = y if not flip_v else TILE_HEIGHT_PX - 1 - y
                pixel_color_ix = 0
                for plane in range(2):
                    pixel_color_ix += ((self.vram.read(tile_base + plane * 8 + y) & (0x1 << (7 - x))) > 0) * (plane + 1)

                if pixel_color_ix > 0:
                    dest[x0 + xx, y0 + yy] = self.hex_palette[palette[pixel_color_ix]]





# cython: profile=True, boundscheck=True, nonecheck=False, language_level=3
import pyximport; pyximport.install()


from .ppu cimport NESPPU
from .carts cimport NESCart
from .system cimport InterruptListener
from .apu cimport NESAPU


######## memory base ##################################################

cdef class MemoryBase:
    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)


######## NES Main Memoroy #############################################

cpdef enum:
    RAM_SIZE = 2048             # 2kB of internal RAM

    # NES Main Memory locations
    RAM_END = 0x2000            # NES main ram to here  (2kb mirrored 4x)
    PPU_END = 0x4000            # PPU registers to here
    APU_END = 0x4018            # APU registers (+OAM DMA reg and controllers) to here
    APU_UNUSED_END = 0x4020     # generally unused APU and I/O functionality
    OAM_DMA_REG = 0x4014        # OAM DMA register address
    CONTROLLER1 = 0x4016        # port for controller (read controller 1 / write both controllers)
    CONTROLLER2 = 0x4017        # port for controller 2 (read only, writes to this port go to the APU)
    CART_START = 0x4020         # start of cartridge address space


cdef class NESMappedRAM(MemoryBase):
    cdef unsigned char ram[RAM_SIZE]
    cdef unsigned char _last_bus
    cdef NESPPU ppu
    cdef NESAPU apu
    cdef NESCart cart
    cdef InterruptListener interrupt_listener
    cdef object controller1, controller2

    ###### functions ##########################
    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)

    cdef void run_oam_dma(self, int page)


######## NES VRAM #####################################################

cdef enum:
    PATTERN_TABLE_SIZE_BYTES = 4096   # provided by the rom

    # the NES VRAM only has 2kb for nametables, but some carts provide for up to 4kb using fixed four-page nametable
    # mirroring.  If we provide enough space for that here, using a mirror pattern of [0, 1, 2, 3] will produce the four
    # page mirroring, and the other mirroring patterns will work as intended, since the addresses will just map to pages
    # 0 and 1 as specified by the mirror pattern supplied by the cartridge.
    XX_NAMETABLES_SIZE_BYTES = 4096
    PALETTE_SIZE_BYTES = 32
    NAMETABLE_LENGTH_BYTES = 1024  # single nametime is this big

    # memory map
    PALETTE_START = 0x3F00
    NAMETABLE_START = 0x2000
    ATTRIBUTE_TABLE_OFFSET = 0x3C0  # offset of the attribute table from the start of the corresponding nametable

cdef class NESVRAM(MemoryBase):
    cdef unsigned char _nametables[XX_NAMETABLES_SIZE_BYTES]
    cdef unsigned char palette_ram[PALETTE_SIZE_BYTES]
    cdef NESCart cart
    cdef int nametable_mirror_pattern[4]

    ###### functions ##########################
    cdef unsigned char read(self, int address)
    cdef void write(self, int address, unsigned char value)



cdef int set_bit(int target, int bit)
cdef int clear_bit(int target, int bit)
cdef int bit_high(int value, int bit)
cdef int bit_low(int value, int bit)
cdef lower_nibble(char value)
cdef upper_nibble(char value)
cdef set_high_byte(int target, int hi_byte)
cdef set_low_byte(int target, int lo_byte)



from collections import namedtuple
from enum import Enum

Instruction = namedtuple("Instruction", ["bytecode", "size_bytes", "cycles"])
NamedInstruction = namedtuple("NamedInstruction", ["bytecode", "size_bytes", "cycles", "name", "mode"])
PyNamedInstruction = namedtuple("NamedInstruction", ["bytecode", "size_bytes", "cycles", "name", "mode", "function"])
InstructionSet = namedtuple("InstructionSet", ["name", "modes"])


class AddressModes(Enum):
    IMMEDIATE = 0
    ZEROPAGE = 1
    ZEROPAGE_X = 2
    ZEROPAGE_Y = 20   # this is only used in LDX and STX in place of ZEROPAGE_X
    ABSOLUTE = 3
    ABSOLUTE_X = 4
    ABSOLUTE_Y = 5
    INDIRECT_X = 6
    INDIRECT_Y = 7
    IMPLIED = 10
    RELATIVE = 11
    ACCUMULATOR = 12
    INDIRECT = 13

INSTRUCTION_SET = {
    "ADC": InstructionSet(name="adc",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x69, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x65, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x75, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x6D, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x7D, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x79, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x61, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x71, size_bytes=2, cycles=5.1),
                                }
                          ),
    "AND": InstructionSet(name="and",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x29, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x25, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x35, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x2D, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x3D, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x39, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x21, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x31, size_bytes=2, cycles=5.1),
                                }
                          ),
    "ASL": InstructionSet(name="asl",
                          modes={AddressModes.ACCUMULATOR:Instruction(bytecode=0x0A, size_bytes=1, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x06, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x16, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x0E, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x1E, size_bytes=3, cycles=7),
                                }
                          ),
    "BCC": InstructionSet(name="bcc",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0x90, size_bytes=2, cycles=2.12)}),
    "BCS": InstructionSet(name="bcs",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0xB0, size_bytes=2, cycles=2.12)}),
    "BEQ": InstructionSet(name="beq",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0xF0, size_bytes=2, cycles=2.12)}),
    "BIT": InstructionSet(name="bit",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x24, size_bytes=2, cycles=3),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x2C, size_bytes=3, cycles=4),
                                }
                          ),
    "BMI": InstructionSet(name="bmi",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0x30, size_bytes=2, cycles=2.12)}),
    "BNE": InstructionSet(name="bne",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0xD0, size_bytes=2, cycles=2.12)}),
    "BPL": InstructionSet(name="bpl",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0x10, size_bytes=2, cycles=2.12)}),
    # note that the size of BRK is slightly controversial, because BRK forces PC+2 to be pushed onto the stack,
    # meaning that the instruction immediately after BRK will never be executed unless there is a workaround applied
    # therefore BRK should usually be followed by NOP, and many assemblers seem to do this
    "BRK": InstructionSet(name="brk",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x00, size_bytes=1, cycles=7)}),
    "BVC": InstructionSet(name="bvc",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0x50, size_bytes=2, cycles=2.12)}),
    "BVS": InstructionSet(name="bvs",
                          modes={AddressModes.RELATIVE:   Instruction(bytecode=0x70, size_bytes=2, cycles=2.12)}),
    "CLC": InstructionSet(name="clc",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x18, size_bytes=1, cycles=2)}),
    "CLD": InstructionSet(name="cld",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xD8, size_bytes=1, cycles=2)}),
    "CLI": InstructionSet(name="cli",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x58, size_bytes=1, cycles=2)}),
    "CLV": InstructionSet(name="clv",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xB8, size_bytes=1, cycles=2)}),
    "CMP": InstructionSet(name="cmp",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xC9, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xC5, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xD5, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xCD, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xDD, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xD9, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xC1, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xD1, size_bytes=2, cycles=5.1),
                                }
                          ),
    "CPX": InstructionSet(name="cpx",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xE0, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xE4, size_bytes=2, cycles=3),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xEC, size_bytes=3, cycles=4),
                                }
                          ),
    "CPY": InstructionSet(name="cpy",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xC0, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xC4, size_bytes=2, cycles=3),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xCC, size_bytes=3, cycles=4),
                                }
                          ),
    "DEC": InstructionSet(name="dec",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0xC6, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xD6, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xCE, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xDE, size_bytes=3, cycles=7),
                                }
                          ),
    "DEX": InstructionSet(name="dex",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xCA, size_bytes=1, cycles=2)}),
    "DEY": InstructionSet(name="dey",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x88, size_bytes=1, cycles=2)}),
    "EOR": InstructionSet(name="eor",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x49, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x45, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x55, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x4D, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x5D, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x59, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x41, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x51, size_bytes=2, cycles=5.1),
                                }
                          ),
    "INC": InstructionSet(name="inc",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0xE6, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xF6, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xEE, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xFE, size_bytes=3, cycles=7),
                                 }
                          ),
    "INX": InstructionSet(name="inx",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xE8, size_bytes=1, cycles=2)}),
    "INY": InstructionSet(name="iny",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xC8, size_bytes=1, cycles=2)}),
    "JMP": InstructionSet(name="jmp",
                          modes={AddressModes.ABSOLUTE:   Instruction(bytecode=0x4C, size_bytes=3, cycles=3),
                                 AddressModes.INDIRECT:   Instruction(bytecode=0x6C, size_bytes=3, cycles=5),
                                 }
                          ),
    "JSR": InstructionSet(name="jsr",
                          modes={AddressModes.ABSOLUTE:   Instruction(bytecode=0x20, size_bytes=3, cycles=6)}),
    "LDA": InstructionSet(name="lda",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xA9, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xA5, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xB5, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xAD, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xBD, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xB9, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xA1, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xB1, size_bytes=2, cycles=5.1),
                                 }
                          ),
    "LDX": InstructionSet(name="ldx",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xA2, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xA6, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_Y: Instruction(bytecode=0xB6, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xAE, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xBE, size_bytes=3, cycles=4.1),
                                 }
                          ),
    "LDY": InstructionSet(name="ldy",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xA0, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xA4, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xB4, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xAC, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xBC, size_bytes=3, cycles=4.1),
                                }
                          ),
    "LSR": InstructionSet(name="lsr",
                          modes={AddressModes.ACCUMULATOR:Instruction(bytecode=0x4A, size_bytes=1, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x46, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x56, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x4E, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x5E, size_bytes=3, cycles=7),
                                 }
                          ),
    "NOP": InstructionSet(name="nop",
                          # the codes after the first code are undocumented NOP codes
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=[0xEA, 0x1A, 0x3A, 0x5A, 0x7A, 0xDA, 0xFA], size_bytes=1, cycles=2)}),
    "ORA": InstructionSet(name="ora",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x09, size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x05, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x15, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x0D, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x1D, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x19, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x01, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x11, size_bytes=2, cycles=5.1),
                                 }
                          ),
    "PHA": InstructionSet(name="pha",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x48, size_bytes=1, cycles=3)}),
    "PHP": InstructionSet(name="php",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x08, size_bytes=1, cycles=3)}),
    "PLA": InstructionSet(name="pla",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x68, size_bytes=1, cycles=4)}),
    "PLP": InstructionSet(name="plp",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x28, size_bytes=1, cycles=4)}),
    "ROL": InstructionSet(name="rol",
                          modes={AddressModes.ACCUMULATOR:Instruction(bytecode=0x2A, size_bytes=1, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x26, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x36, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x2E, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x3E, size_bytes=3, cycles=7),
                                 }
                          ),
    "ROR": InstructionSet(name="ror",
                          modes={AddressModes.ACCUMULATOR:Instruction(bytecode=0x6A, size_bytes=1, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0x66, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x76, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x6E, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x7E, size_bytes=3, cycles=7),
                                 }
                          ),
    "RTI": InstructionSet(name="rti",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x40, size_bytes=1, cycles=6)}),
    "RTS": InstructionSet(name="rts",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x60, size_bytes=1, cycles=6)}),
    "SBC": InstructionSet(name="sbc",
                          # the second code for the immediate version is an illegal opcode
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=[0xE9, 0xEB], size_bytes=2, cycles=2),
                                 AddressModes.ZEROPAGE:   Instruction(bytecode=0xE5, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xF5, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xED, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xFD, size_bytes=3, cycles=4.1),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xF9, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xE1, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xF1, size_bytes=2, cycles=5.1),
                                 }
                          ),
    "SEC": InstructionSet(name="sec",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x38, size_bytes=1, cycles=2)}),
    "SED": InstructionSet(name="sed",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xF8, size_bytes=1, cycles=2)}),
    "SEI": InstructionSet(name="sei",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x78, size_bytes=1, cycles=2)}),
    "STA": InstructionSet(name="sta",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x85, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x95, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x8D, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x9D, size_bytes=3, cycles=5),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x99, size_bytes=3, cycles=5),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x81, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x91, size_bytes=2, cycles=6),
                                 }
                          ),
    "STX": InstructionSet(name="stx",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x86, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_Y: Instruction(bytecode=0x96, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x8E, size_bytes=3, cycles=4),
                                 }
                          ),
    "STY": InstructionSet(name="sty",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x84, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x94, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x8C, size_bytes=3, cycles=4),
                                 }
                          ),
    "TAX": InstructionSet(name="tax",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xAA, size_bytes=1, cycles=2)}),
    "TAY": InstructionSet(name="tay",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xA8, size_bytes=1, cycles=2)}),
    "TSX": InstructionSet(name="tsx",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0xBA, size_bytes=1, cycles=2)}),
    "TXA": InstructionSet(name="txa",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x8A, size_bytes=1, cycles=2)}),
    "TXS": InstructionSet(name="txs",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x9A, size_bytes=1, cycles=2)}),
    "TYA": InstructionSet(name="tya",
                          modes={AddressModes.IMPLIED:    Instruction(bytecode=0x98, size_bytes=1, cycles=2)}),

    # Undocumented instructions below here
    # naming convention is from http://nesdev.com/undocumented_opcodes.txt, several others exist
    "DOP": InstructionSet(name="dop",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=[0x04, 0x14, 0x44, 0x64], size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=[0x14, 0x34, 0x54, 0x74, 0xD4, 0xF4], size_bytes=2, cycles=4),
                                 AddressModes.IMMEDIATE:  Instruction(bytecode=[0x80, 0x82, 0x89, 0xC2, 0xE2], size_bytes=2, cycles=2),
                                 }
                          ),
    "TOP": InstructionSet(name="top",
                          modes={AddressModes.ABSOLUTE:   Instruction(bytecode=0x0C, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=[0x1C, 0x3C, 0x5C, 0x7C, 0xDC, 0xFC], size_bytes=3, cycles=4.1),
                                 }
                          ),
    "KIL": InstructionSet(name="kil",
                      modes={AddressModes.IMPLIED:        Instruction(bytecode=[0x02, 0x12, 0x22, 0x32, 0x42, 0x52, 0x62, 0x72, 0x92, 0xB2, 0xD2, 0xF2], size_bytes=1, cycles=1),
                             }
                         ),
    "LAX": InstructionSet(name="lax",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0xA7, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_Y: Instruction(bytecode=0xB7, size_bytes=2, cycles=4),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xAF, size_bytes=3, cycles=4),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xBF, size_bytes=3, cycles=4.1),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xA3, size_bytes=2, cycles=6),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xB3, size_bytes=2, cycles=5.1),
                                 }
                          ),
    "AAX": InstructionSet(name="aax",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x87, size_bytes=2, cycles=3),
                                 AddressModes.ZEROPAGE_Y: Instruction(bytecode=0x97, size_bytes=2, cycles=4),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x83, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x8F, size_bytes=3, cycles=4),
                                 }
                          ),
    "DCP": InstructionSet(name="dcp",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0xC7, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xD7, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xCF, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xDF, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xDB, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xC3, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xD3, size_bytes=2, cycles=8),
                                 }
                          ),
    "ISC": InstructionSet(name="isc",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0xE7, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0xF7, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0xEF, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0xFF, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xFB, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0xE3, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0xF3, size_bytes=2, cycles=8),
                                 }
                          ),
    "SLO": InstructionSet(name="slo",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x07, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x17, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x0F, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x1F, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x1B, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x03, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x13, size_bytes=2, cycles=8),
                                 }
                          ),
    "RLA": InstructionSet(name="rla",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x27, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x37, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x2F, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x3F, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x3B, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x23, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x33, size_bytes=2, cycles=8),
                                 }
                          ),
    "SRE": InstructionSet(name="sre",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x47, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x57, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x4F, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x5F, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x5B, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x43, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x53, size_bytes=2, cycles=8),
                                 }
                          ),
    "RRA": InstructionSet(name="rra",
                          modes={AddressModes.ZEROPAGE:   Instruction(bytecode=0x67, size_bytes=2, cycles=5),
                                 AddressModes.ZEROPAGE_X: Instruction(bytecode=0x77, size_bytes=2, cycles=6),
                                 AddressModes.ABSOLUTE:   Instruction(bytecode=0x6F, size_bytes=3, cycles=6),
                                 AddressModes.ABSOLUTE_X: Instruction(bytecode=0x7F, size_bytes=3, cycles=7),
                                 AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x7B, size_bytes=3, cycles=7),
                                 AddressModes.INDIRECT_X: Instruction(bytecode=0x63, size_bytes=2, cycles=8),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x73, size_bytes=2, cycles=8),
                                 }
                          ),
    # Implemented but untested
    "ARR": InstructionSet(name="arr",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x6B, size_bytes=2, cycles=2),
                                 }
                          ),
    "ASR": InstructionSet(name="asr",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x4B, size_bytes=2, cycles=2),
                                 }
                          ),
    "ATX": InstructionSet(name="atx",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xAB, size_bytes=2, cycles=2),
                                 }
                          ),
    "AAC": InstructionSet(name="aac",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=[0x0B, 0x2B], size_bytes=2, cycles=2),
                                 }
                          ),
    "AXA": InstructionSet(name="axa",
                          modes={AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x9F, size_bytes=3, cycles=5),
                                 AddressModes.INDIRECT_Y: Instruction(bytecode=0x93, size_bytes=2, cycles=6),
                                 }
                          ),
    "AXS": InstructionSet(name="axs",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0xCB, size_bytes=2, cycles=2),
                                 }
                          ),
    "LAR": InstructionSet(name="lar",
                          modes={AddressModes.ABSOLUTE_Y: Instruction(bytecode=0xBB, size_bytes=3, cycles=4.1),
                                 }
                          ),
    "SXA": InstructionSet(name="sxa",
                          modes={AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x9E, size_bytes=3, cycles=5),
                                 }
                          ),
    "SYA": InstructionSet(name="sya",
                          modes={AddressModes.ABSOLUTE_X: Instruction(bytecode=0x9C, size_bytes=3, cycles=5),
                                 }
                          ),
    "XAA": InstructionSet(name="xaa",
                          modes={AddressModes.IMMEDIATE:  Instruction(bytecode=0x8B, size_bytes=2, cycles=2),
                                 }
                          ),
    "XAS": InstructionSet(name="xas",
                          modes={AddressModes.ABSOLUTE_Y: Instruction(bytecode=0x9B, size_bytes=3, cycles=5),
                                 }
                          ),
}